"""
generate performances before/after US election dates and conditional on the divided/unified goverment
"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
from pandas.tseries.holiday import AbstractHolidayCalendar, Holiday
from dateutil.relativedelta import TU
from bbg_fetch import fetch_field_timeseries_per_tickers
from typing import Tuple, Literal, Dict, Optional
from enum import Enum

# this table is generated using data in https://stevesnotes.substack.com/p/note-1-gridlock
# year is matched to the election data
US_ELECTION_RESULTS = {1968: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1970: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1972: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1974: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1976: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Democrat', Divided_Unified='Unified'),
                       1978: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Democrat', Divided_Unified='Unified'),
                       1980: dict(Senate_Majority='Republicans', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1982: dict(Senate_Majority='Republicans', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1984: dict(Senate_Majority='Republicans', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1986: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1988: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1990: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       1992: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Democrat', Divided_Unified='Unified'),
                       1994: dict(Senate_Majority='Republicans', House_Majority='Republicans', President='Democrat', Divided_Unified='Divided'),
                       1996: dict(Senate_Majority='Republicans', House_Majority='Republicans', President='Democrat', Divided_Unified='Divided'),
                       1998: dict(Senate_Majority='Republicans', House_Majority='Republicans', President='Democrat', Divided_Unified='Divided'),
                       2000: dict(Senate_Majority='Democrats', House_Majority='Republicans', President='Republican', Divided_Unified='Divided'),
                       2002: dict(Senate_Majority='Republicans', House_Majority='Republicans', President='Republican', Divided_Unified='Unified'),
                       2004: dict(Senate_Majority='Republicans', House_Majority='Republicans', President='Republican', Divided_Unified='Unified'),
                       2006: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       2008: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Democrat', Divided_Unified='Unified'),
                       2010: dict(Senate_Majority='Democrats', House_Majority='Republicans', President='Democrat', Divided_Unified='Divided'),
                       2012: dict(Senate_Majority='Democrats', House_Majority='Republicans', President='Democrat', Divided_Unified='Divided'),
                       2014: dict(Senate_Majority='Republicans', House_Majority='Republicans', President='Democrat', Divided_Unified='Divided'),
                       2016: dict(Senate_Majority='Republicans', House_Majority='Republicans', President='Republican', Divided_Unified='Unified'),
                       2018: dict(Senate_Majority='Republicans', House_Majority='Democrats', President='Republican', Divided_Unified='Divided'),
                       2020: dict(Senate_Majority='Democrats', House_Majority='Democrats', President='Democrat', Divided_Unified='Unified'),
                       2022: dict(Senate_Majority='Democrats', House_Majority='Republicans', President='Democrat', Divided_Unified='Divided')
                       }


def generate_us_election_dates(start_date: pd.Timestamp = pd.Timestamp('01Jan1950'),
                               end_date: pd.Timestamp = pd.Timestamp('01Jan2023')
                               ) -> pd.DatetimeIndex:
    """
    see https://stackoverflow.com/questions/34708626/pandas-holiday-calendar-rule-for-us-election-day
    """
    def election_observance(dt):
        if not dt.year % 4 == 0:  # every fourth year
            return None
        else:
            return dt + pd.DateOffset(weekday=TU(1))

    class USElectionCalendar(AbstractHolidayCalendar):
        """
        Federal Presidential  and Congressional election day.
        Tuesday following the first Monday, 2 to 8 November every two even numbered years.
        Election Days can only occur from November 2nd through 8th inclusive.
        """
        rules = [
            Holiday('Election Day', month=11, day=2, observance=election_observance)
        ]

    cal = USElectionCalendar()
    dates = cal.holidays(start_date, end_date, return_name=False)
    return dates


def generate_before_after_dates(dates: pd.DatetimeIndex,
                                period: int = 150
                                ) -> Dict[pd.Timestamp, Tuple[pd.Timestamp, pd.Timestamp]]:
    """
    period is measured in business days
    """
    before_after_dates = {}
    for date in dates:
        before_after_dates[date] = (date-pd.DateOffset(days=period), date+pd.DateOffset(days=period))
    return before_after_dates


def create_before_after_performances(price: pd.Series,
                                     before_after_dates: Dict[pd.Timestamp, Tuple[pd.Timestamp, pd.Timestamp]]
                                     ) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    price should be at 'B' frequency
    """
    before_prices = {}
    after_prices = {}
    for date, before_after_date in before_after_dates.items():
        before_price = price.loc[before_after_date[0]:date]
        after_price = price.loc[date:before_after_date[1]]
        # normalise
        if not before_price.empty:
            before_price = before_price / before_price.iloc[-1]
            before_price.index = np.arange(-len(before_price.index)+1, 1, 1)
            before_prices[date.year] = before_price

        if not after_price.empty:
            after_price = after_price / after_price.iloc[0]
            after_price.index = np.arange(0, len(after_price.index))
            after_prices[date.year] = after_price
    before_prices = pd.DataFrame.from_dict(before_prices, orient='columns')
    after_prices = pd.DataFrame.from_dict(after_prices, orient='columns')
    return before_prices, after_prices


def merge_before_after_prices(before_prices: pd.DataFrame,
                              after_prices: pd.DataFrame
                              ) -> pd.DataFrame:
    joint_price = pd.concat([before_prices, after_prices.iloc[1:, :]], axis=0)
    joint_price = joint_price[joint_price.columns[::-1]]
    return joint_price


def plot_before_after_prices(joint_prices: pd.DataFrame,
                             add_average: bool = True,
                             xlabel: str = 'Days before/after election date',
                             ylabel: str = 'Normalised performance',
                             title: str = None,
                             is_norm: bool = True,
                             ax: plt.Subplot = None
                             ) -> Optional[plt.Figure]:

    if add_average:
        joint_prices.insert(0, column='Average', value=joint_prices.mean(1))
    if is_norm:
        joint_prices = joint_prices - 1.0
        yvar_format='{:,.0%}'
    else:
        yvar_format='{:,.2f}',
    n = len(joint_prices.columns)
    colors = qis.get_cmap_colors(n=n, cmap='tab20')
    linestyles = ['dotted'] * n
    if add_average:
        colors[0] = 'blue'
        linestyles[0] = '-'

    if ax is None:
        with sns.axes_style("whitegrid"):
            fig, ax = plt.subplots(1, 1, figsize=(8, 8))
    else:
        fig = None
    qis.plot_line(df=joint_prices,
                  xlabel=xlabel,
                  ylabel=ylabel,
                  title=title,
                  colors=colors,
                  linestyles=linestyles,
                  yvar_format=yvar_format,
                  framealpha=0.9,
                  ax=ax)
    ax.axvline(0, color='lightsalmon', lw=1, alpha=0.5)  # horizontal
    return fig


def plot_unconditional_performances(price: pd.Series, title: str = None, ax: plt.Subplot = None) -> plt.Figure:
    dates = generate_us_election_dates()
    before_after_dates = generate_before_after_dates(dates)
    before_prices, after_prices = create_before_after_performances(price=price,
                                                                   before_after_dates=before_after_dates)
    joint_prices = merge_before_after_prices(before_prices=before_prices, after_prices=after_prices)
    fig = plot_before_after_prices(joint_prices=joint_prices, title=title, ax=ax)
    return fig


def compute_conditional_performances(
        price: pd.Series,
        classifier: Literal['Senate_Majority', 'House_Majority', 'President', 'Divided_Unified'] = 'Divided_Unified'
        ) -> Dict[str, pd.DataFrame]:
    """
    compute before / after performances conditional on classifier
    """
    dates = generate_us_election_dates()
    before_after_dates = generate_before_after_dates(dates)
    before_prices, after_prices = create_before_after_performances(price=price,
                                                                   before_after_dates=before_after_dates)
    joint_price = merge_before_after_prices(before_prices=before_prices, after_prices=after_prices)
    t_joint_price = joint_price.T
    # add classifiers
    election_results = pd.DataFrame.from_dict(US_ELECTION_RESULTS, orient='index').loc[t_joint_price.index]
    t_joint_price[classifier] = election_results[classifier]
    t_joint_price = t_joint_price.groupby(classifier)
    dfs = {}
    for group, prices in t_joint_price:
        dfs[str(group)] = prices.drop(classifier, axis=1).T
    return dfs


def plot_conditional_performances(dfs: Dict[str, pd.DataFrame], title: str) -> plt.Figure:
    with sns.axes_style("whitegrid"):
        fig, axs = plt.subplots(1, len(dfs.keys()), figsize=(14, 8))
        qis.set_suptitle(fig, title=title)
        for idx, (group, joint_prices) in enumerate(dfs.items()):
            plot_before_after_prices(joint_prices=joint_prices, title=group, ax=axs[idx])
        qis.align_y_limits_axs(axs)
    return fig


class UnitTests(Enum):
    ELECTION_RESULTS = 1
    ELECTION_DATES = 2
    UNCONDITIONAL_PERFORMANCES = 3
    CONDITIONAL_PERFORMANCES = 4


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.ELECTION_RESULTS:
        df = pd.DataFrame.from_dict(US_ELECTION_RESULTS, orient='index')
        print(df)

    elif unit_test == UnitTests.ELECTION_DATES:
        dates = generate_us_election_dates()
        print(dates)

        before_after_dates = generate_before_after_dates(dates)
        print(before_after_dates)

    elif unit_test == UnitTests.UNCONDITIONAL_PERFORMANCES:
        tickers = {'SPX Index': 'S&P 500',
                   'DXY Curncy': 'DXY',
                   'TY1 Comdty': 'UST 10y bond future'}
        prices = fetch_field_timeseries_per_tickers(tickers=tickers, freq='B').ffill()
        figs = []
        for asset in prices.columns:
            fig = plot_unconditional_performances(price=prices[asset].dropna(), title=f"{asset} performance")
            figs.append(fig)
        qis.save_figs_to_pdf(figs, file_name='election_perf_unconditional', local_path=qis.get_output_path())

    elif unit_test == UnitTests.CONDITIONAL_PERFORMANCES:
        tickers = {'SPX Index': 'S&P 500',
                   'DXY Curncy': 'DXY',
                   'TY1 Comdty': 'UST 10y bond future'}
        tickers = {'NKY Index': 'Nikkei 225',
                   'HSI Index': 'HSI',
                   'SXXP Index': 'SXXP',
                   'SPX Index': 'S&P 500',
                   'DXY Curncy': 'DXY',
                   'TY1 Comdty': 'UST 10y bond future'
                   }
        prices = fetch_field_timeseries_per_tickers(tickers=tickers, freq='B',
                                                    start_date=pd.Timestamp('01Jan1968')).ffill()
        print(prices)
        figs = []
        dfss = {}
        for asset in prices.columns:
            price = prices[asset].dropna()
            dfs = compute_conditional_performances(price=price)
            fig = plot_conditional_performances(dfs=dfs, title=f"{asset} performance conditional on US Government")
            for key, df in dfs.items():
                dfss[f"{asset}-{key}"] = df - 1.0
            figs.append(fig)
        qis.save_figs_to_pdf(figs, file_name='perf_conditional_on_government', local_path=qis.get_output_path())
        qis.save_df_to_excel(data=dfss, file_name='perf_conditional_on_government', local_path=qis.get_output_path())
    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.CONDITIONAL_PERFORMANCES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# packages
import numpy as np
import pandas as pd
from functools import partial
from enum import Enum
from pandas.core.dtypes.common import is_datetime64_any_dtype as is_datetime
from pandas.api.types import is_string_dtype
from tabulate import tabulate
from typing import List, Optional, Union, Dict

import qis
from qis.utils.dates import DATE_FORMAT

EMPTY_NUM = ' '


# Formatting string
def get_fmt_str(x, fill):
    return '{message: >{fill}}'.format(message=x, fill=fill)


def float_to_str(x: float,
                 var_format: str = '{:.2f}',
                 is_exclude_nans: bool = True
                 ) -> str:
    """
    float to str dendent on nan and var_format
    use as pd.Series().apply(lambda x: float_to_str(x, var_format))
    """
    if is_exclude_nans and isinstance(x, float) and np.isnan(x):
        return EMPTY_NUM
    elif isinstance(x, str):
        return x
    else:
        if not isinstance(x, float):
            raise ValueError(f"x must ne float not type={type(x)}")
        return var_format.format(x)


def str_to_float(x: str) -> float:
    """
    str to float
    use as pd.Series().apply(lambda x: str_to_float(x))
    """
    if isinstance(x, float) and np.isnan(x):  # just in case
        x = np.nan
    elif isinstance(x, str) and x == EMPTY_NUM:
        x = np.nan
    else:
        try:
            x = float(x.replace(',', '').replace('%', ''))
        except:  # the string contains words
            x = np.nan
    return x


def date_to_str(x: pd.Timestamp, var_format: str = DATE_FORMAT) -> str:
    """
    float to str dendent on nan and var_format
    use as pd.Series().apply(lambda x: float_to_str(x, var_format))
    """
    if x is None or x is pd.NaT or pd.notnull(x)==False:
        x_str = EMPTY_NUM
    else:  # var_format.format(x) does not work for date str
        #if isinstance(x, pd.Timestamp):
        #    x_str = x.strftime(var_format)
        #else:
        x_str = pd.Timestamp(x).strftime(var_format)
    return x_str


def series_to_str(ds: pd.Series,
                  var_format: str = '{:.2f}',
                  is_exclude_nans: bool = True
                  ) -> pd.Series:
    """
    pd.Series to string using float_to_str
    """
    if not isinstance(ds, pd.Series):
        print(ds)
        raise TypeError(f"data type= {type(ds)} must be pd.Series")

    if is_string_dtype(ds):
        pd_series = ds.copy()
    elif is_datetime(ds):
        # pycharm geeting wrong typecheck https://youtrack.jetbrains.com/issue/PY-43841
        pd_series = ds.apply(lambda x: date_to_str(x, var_format))
    else:
        pd_series = ds.apply(lambda x: float_to_str(x, var_format, is_exclude_nans))

    if is_exclude_nans:
        pd_series[pd_series.isna()] = EMPTY_NUM

    return pd_series


def series_to_date_str(ds: pd.Series,
                       var_format: str = DATE_FORMAT,
                       is_exclude_nans: bool = True
                       ) -> pd.Series:
    """
    pd.Series to string using float_to_str
    """
    if not isinstance(ds, pd.Series):
        print(ds)
        raise TypeError(f"data type= {type(ds)} must be pd.Series")

    pd_series = ds.apply(lambda x: date_to_str(x, var_format))
    if is_exclude_nans:
        pd_series[pd_series.isna()] = EMPTY_NUM
    return pd_series


def series_to_numeric(ds: pd.Series) -> np.ndarray:
    if ds.dtypes == np.float64:
        a = ds.to_numpy()
    elif ds.dtypes == str or isinstance(ds.dtypes, object):
        a = ds.apply(lambda x: str_to_float(x)).to_numpy()
    else:
        raise ValueError(f"unsupported value type: {ds.dtypes}")
    return a


def df_to_numeric(df: pd.DataFrame) -> np.ndarray:
    a = df.apply(lambda x: series_to_numeric(x)).to_numpy()
    return a


def df_to_str(df: pd.DataFrame,
              var_format: Optional[str] = '{:.2f}',
              var_formats: Union[List[Optional[str]], Dict[str, str]] = None,  # specific for each column
              is_exclude_nans: bool = True
              ) -> pd.DataFrame:
    """
    pd.DataFrame to string using float_to_str
    """
    if var_formats is not None:

        if isinstance(var_formats, list):
            if not len(var_formats) == len(df.columns):
                raise ValueError(f"match len of var_formats {var_formats} with {df.columns}")
        elif isinstance(var_formats, dict):

            if var_format is not None:  # use var_format as pivot
                var_formats_missing = {x: var_format for x in df.columns}
                var_formats = qis.update_kwargs(var_formats_missing, var_formats)

            var_formats_ = []
            for column in df.columns:
                if column in var_formats.keys():
                    var_formats_.append(var_formats[column])
                else:
                    var_formats_.append(None)
            var_formats = var_formats_
        else:
            raise ValueError(f"{var_formats} is not supported")

    else:
        var_formats = [var_format]*len(df.columns)
    df = df.copy()
    for column, var_format in zip(df.columns, var_formats):
        if var_format is not None:
            df[column] = series_to_str(ds=df[column], var_format=var_format, is_exclude_nans=is_exclude_nans)
    return df


def timeseries_df_to_str(df: pd.DataFrame,
                         freq: Optional[str] = 'QE',
                         date_format: str = '%b-%y',
                         var_format: str = '{:.0%}',
                         var_formats: List[str] = None,  # specific for each column
                         transpose: bool = True
                         ) -> pd.DataFrame:
    """
    time series df to table str
    """
    if freq is not None:
        df = df.resample(freq).last()
    df.index = df.index.strftime(date_format)
    if transpose:
        df = df.T
    df = df_to_str(df, var_format=var_format, var_formats=var_formats)
    return df


def df_with_ci_to_str(df: pd.DataFrame,
                      df_ci: pd.DataFrame,
                      var_format: str = '{:.2f}',
                      is_exclude_nans: bool = True,
                      sep: str = u"\u00B1"
                      ) -> pd.DataFrame:
    """
    pd.DataFrame to string with ci intervals
    """
    df_out = pd.DataFrame(index=df.index, columns=df.columns)
    for column in df.columns:
        val = series_to_str(ds=df[column], var_format=var_format, is_exclude_nans=is_exclude_nans)
        ci = series_to_str(ds=df_ci[column], var_format=var_format, is_exclude_nans=is_exclude_nans)
        df_out[column] = join_str_series(ds1=val, ds2=ci, sep=sep)
    return df_out


def join_str_series(ds1: pd.Series, ds2: pd.Series, sep: str = u"\u00B1") -> np.ndarray:
    """
    joint str series with default separator = +-
    """
    out = np.empty(len(ds1.index), dtype=object)
    for idx, (n1, n2) in enumerate(zip(ds1.to_numpy(), ds2.to_numpy())):
        out[idx] = f"{n1}{sep}{n2}"
    return out


def df_all_to_str(df: pd.DataFrame, index_name: str = '') -> str:
    # Max character length per column
    df.index.name = index_name
    df = df.reset_index()
    s = df.astype(str).agg(lambda x: x.str.len()).max()

    pad = 10  # How many spaces between
    fmts = {}
    for idx, c_len in s.items():
        # Deal with MultIndex tuples or simple string labels.
        if isinstance(idx, tuple):
            lab_len = max([len(str(x)) for x in idx])
        else:
            lab_len = len(str(idx))

        fill = max(lab_len, c_len) + pad - 1
        fmts[idx] = partial(get_fmt_str, fill=fill)
    df_str = df.apply(fmts)
    stats_str = tabulate(df_str, showindex=False, floatfmt='.2f', headers=df.columns)
    return stats_str


def series_values_to_str(ds: pd.Series, include_index: bool = True) -> str:
    data_dict = ds.to_dict()
    data_str = ""
    for k, v in data_dict.items():
        if include_index:
            data_str += f"{k}: {v}, "
        else:
            data_str += f"{v}, "
    return data_str


def df_index_to_str(df: pd.DataFrame,
                    freq: str = 'QE',
                    data_str: str = 'Q%q-%y'
                    ) -> pd.DataFrame:
    df.index = pd.PeriodIndex(pd.to_datetime(df.index).date, freq=freq).strftime(data_str)
    return df


def idx_to_alphabet(idx: int = 1, capitalise: bool = True) -> str:
    """
    map index to alphabet character
    """
    if capitalise:
        return chr(ord('@') + idx)
    else:
        return chr(ord('`') + idx)


class UnitTests(Enum):
    DF_TO_STR = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.DF_TO_STR:
        df = pd.DataFrame({
            "c1": ("a", "bb", "ccc", "dddd", "eeeeee"),
            "c2": (11, 22, 33, 44, 55),
            "a3235235235": [1, 2, 3, 4, 5]
        })
        print(df)

        fmts = df_all_to_str(df)
        print(fmts)

        stats_str = tabulate(df, showindex=True, floatfmt='.2f', headers=df.columns)
        print(stats_str)


if __name__ == '__main__':

    unit_test = UnitTests.DF_TO_STR

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
common pandas operations
"""
# packages
import warnings
import numpy as np
import pandas as pd
from scipy import stats
from typing import Union, List, Optional, Dict, Tuple, Type, Any, Literal
from enum import Enum
# qis
import qis.utils.np_ops as npo
import qis.utils.struct_ops as sop


def df_zero_like(df: pd.DataFrame) -> pd.DataFrame:
    return pd.DataFrame(data=np.zeros(df.shape), index=df.index, columns=df.columns)


def df_ones_like(df: pd.DataFrame) -> pd.DataFrame:
    return pd.DataFrame(data=np.ones(df.shape), index=df.index, columns=df.columns)


def df_indicator_like(df: pd.DataFrame, type: Type = bool) -> pd.DataFrame:
    """
    return df_indicator = True is set is not None and False otherwise
    """
    data = (df.isna() == False).astype(type)
    return pd.DataFrame(data=data, index=df.index, columns=df.columns)


def dfs_indicators(dfs: List[pd.DataFrame], type: Type = bool) -> pd.DataFrame:
    """
    return df_indicator = True if list dfs is not none
    nb: need to check that dfs are aligned
    """
    def to_indicator(df: pd.DataFrame) -> np.ndarray:
        return np.where(np.isnan(df.to_numpy(dtype=np.float64)), False, True)
    indicators = to_indicator(df=dfs[0])
    for df in dfs[1:]:
        indicators = np.logical_and(indicators, to_indicator(df=df))
    return pd.DataFrame(data=indicators, index=dfs[0].index, columns=dfs[0].columns).astype(type)


def df_joint_indicator(indicator1: pd.DataFrame,
                       indicator2: Union[np.ndarray, pd.DataFrame],
                       to_float: bool = False
                       ) -> pd.DataFrame:
    """
    return df_indicator = True if indicator1 is True and indicator2 is True
    """
    if isinstance(indicator2, pd.DataFrame):
        np_indicator2 = indicator2.to_numpy(dtype=bool)
    elif isinstance(indicator2, np.ndarray):
        np_indicator2 = indicator2.astype(bool)
    else:
        raise ValueError(f"unsupported type {type(indicator2)}")

    data = np.logical_and(indicator1.to_numpy(dtype=bool), np_indicator2)

    if to_float:
        data = data.astype(np.float64)

    return pd.DataFrame(data=data, index=indicator1.index, columns=indicator1.columns)


def norm_df_by_ax_mean(data: Union[np.ndarray, pd.DataFrame],
                       proportiontocut: Optional[float] = None,  # for treaming mean,
                       is_zero_mean: bool = True,
                       axis: int = 1
                       ) -> pd.DataFrame:
    """
    to unit norm data
    axis: 0: time-series row-wise mean; 1: cross-sectional columns-wise mean
    """

    if isinstance(data, pd.DataFrame):
        a = data.to_numpy()
    else:
        a = data

    if is_zero_mean:
        if proportiontocut is not None:
            mean = npo.np_array_to_df_columns(a=stats.trim_mean(a=a,
                                                                proportiontocut=proportiontocut, axis=axis),
                                              ncols=len(data.columns))
        else:
            mean = np.nanmean(a=a, axis=axis, keepdims=True)
        a = a - mean

    std = np.nanstd(a=a, axis=axis, ddof=1, keepdims=True)
    norm_data = np.divide(a, std, where=np.greater(std, 0.0))

    if isinstance(data, pd.DataFrame):
        norm_data = pd.DataFrame(data=norm_data, columns=data.columns, index=data.index)

    return norm_data


def get_first_before_nonnan_index(df: Union[pd.Series, pd.DataFrame],
                                  return_index_for_all_nans: int = -1  # last = -1, first = 0
                                  ) -> Union[pd.Timestamp, List[pd.Timestamp]]:
    """
    return first index if no nans in pandas
    """
    if isinstance(df, pd.DataFrame):
        first_before_nonnan_index = []
        for column in df:
            column_data = df[column]
            non_nan_data = column_data.loc[np.isnan(column_data.to_numpy()) == False]
            if not non_nan_data.empty:  # returns first index of nan
                fist_nonnan_index = non_nan_data.index[0]
            else:  # all values are nans so return the last index of original data
                fist_nonnan_index = df.index[return_index_for_all_nans]

            first_before_nonnan_index.append(fist_nonnan_index)

    elif isinstance(df, pd.Series):
        non_nan_data = df.loc[np.isnan(df.to_numpy()) == False]
        if not non_nan_data.empty:
            first_before_nonnan_index = non_nan_data.index[0]
        else:
            first_before_nonnan_index = df.index[return_index_for_all_nans]
    else:
        raise ValueError(f"unsoported data type = {type(df)}")

    return first_before_nonnan_index


def drop_first_nan_data(df: Union[pd.Series, pd.DataFrame],
                        is_oldest: bool = True  # the olderst, the eldest (all non nnans)
                        ) -> Union[pd.Series, pd.DataFrame]:
    """
    drop data before first nonnan either at max or at min for pandas
    the rest of data can still contain occasional nans
    """
    first_nonnan_index = get_first_before_nonnan_index(df=df)

    if isinstance(df, pd.DataFrame):
        if is_oldest:
            joint_start = min(first_nonnan_index)
        else:
            joint_start = max(first_nonnan_index)
        new_data = df[joint_start:].copy()
    else:
        new_data = df.loc[first_nonnan_index:].copy()

    return new_data


def get_first_last_nonnan_index(df: Union[pd.Series, pd.DataFrame],
                                is_first: bool = True
                                ) -> Union[pd.Timestamp, np.ndarray]:
    """
    for given time series df or series:
    find the first date of non-nan value if  is_first=True
    find the last date of non-nan value
    return pd.Timestamp or np.nan
    """
    def get_series_non_nan(ds: pd.Series) -> pd.Timestamp:
        null_ind = ds.isnull()
        good_ind = null_ind == False
        if np.all(good_ind):
            if is_first:  # no nans nans return first:
                fist_nonnan_index = ds.index[0]
            else:  # last
                fist_nonnan_index = ds.index[-1]
        elif np.all(null_ind):
            if is_first:  # all nans return last:
                fist_nonnan_index = ds.index[-1]
            else:
                fist_nonnan_index = ds.index[0]
        else:
            if is_first:
                fist_nonnan_index = ds[good_ind].index[0]
            else:
                fist_nonnan_index = ds[good_ind].index[-1]
        return fist_nonnan_index

    if isinstance(df, pd.Series):
        fist_nonnan_index = get_series_non_nan(df)
    elif isinstance(df, pd.DataFrame):
        values = []
        for column in df:
            values.append(get_series_non_nan(df[column]))
        fist_nonnan_index = np.array(values)
    else:
        raise ValueError(f"unsupported data type = {type(df)}")

    return fist_nonnan_index


def compute_nans_zeros_ratio_after_first_non_nan(df: Union[pd.Series, pd.DataFrame],
                                                 zero_cutoff: float = 1e-12
                                                 ) -> Tuple[np.ndarray, np.ndarray]:
    """
   compute ratio  of missing data
    """
    if isinstance(df, pd.Series):
        df = df.to_frame()

    missing = []
    zeros = []
    for column in df:
        column_data = df[column]
        nonnan_cond = column_data.isnull() == False
        fist_nonnan_index = column_data.loc[nonnan_cond].index[0]
        after_data = column_data.loc[fist_nonnan_index:]
        missing.append(after_data.isnull().sum() / len(after_data))
        zeros.append(after_data.abs().lt(zero_cutoff).sum() / len(after_data))

    missing_ratio = np.array(missing)
    zeros_ratio = np.array(zeros)

    return missing_ratio, zeros_ratio


def get_first_nonnan_values(df: Union[np.ndarray, pd.Series, pd.DataFrame]) -> Union[np.ndarray, float]:

    if isinstance(df, pd.DataFrame) or isinstance(df, pd.Series):
        if df.empty:
            raise ValueError(f"data is empty:\n {df}")

    def get_non_nan_values_series(sdata: pd.Series) -> float:
        x0 = sdata.iloc[0]
        if not np.isnan(x0):
            values = x0
        else:
            nonnan_column_data = sdata[~sdata.isnull()]
            if nonnan_column_data.empty:
                warnings.warn(f"all nans in {df}")
                values = np.nan
            else:
                values = nonnan_column_data.iloc[0]
        return values

    if isinstance(df, pd.DataFrame):
        x0 = df.iloc[0, :].to_numpy(float)
        if np.all(np.isnan(x0) == False):  # first entries are non nan -> most expected
            values = x0
        else:
            values = []
            for column in df:
                values.append(get_non_nan_values_series(sdata=df[column]))
            values = np.array(values)

    elif isinstance(df, pd.Series):
            values = get_non_nan_values_series(sdata=df)

    elif isinstance(df, np.ndarray):
        x0 = df[0, :]
        if np.all(np.isnan(x0) == False):  # first entries are non nan -> most expected
            values = x0
        else:
            values = []
            for idx in np.arange(df.shape[1]):
                values.append(get_non_nan_values_series(sdata=pd.Series(df[:, idx])))
            values = np.array(values)

    else:
        raise ValueError(f"unsupported data type = {type(df)}")

    return values


def get_last_nonnan_values(df: Union[pd.Series, pd.DataFrame]) -> Union[np.ndarray, float]:
    if df.empty:
        raise ValueError(f"data is empty:\n {df}")

    def get_non_nan_values_series(ds: pd.Series) -> float:
        x1 = ds.iloc[-1]
        if not pd.isna(x1):
            values = x1
        else:
            nonnan_column_data = ds[~ds.isnull()]
            if nonnan_column_data.empty:
                # print(f"in get_last_non_nan_values: all nans in {ds}")
                values = np.nan
            else:
                values = nonnan_column_data.iloc[-1]
        return values

    if isinstance(df, pd.DataFrame):
        x1 = df.iloc[-1, :].to_numpy()
        if np.all(pd.isna(x1) == False):  # first entries are non nan -> most expected
            values = x1
        else:
            values = []
            for column in df:
                values.append(get_non_nan_values_series(ds=df[column]))
            values = np.array(values)

    elif isinstance(df, pd.Series):
        values = get_non_nan_values_series(ds=df)

    else:
        raise ValueError(f"unsupported data type = {type(df)}")

    return values


def get_last_nonnan(df: Union[pd.Series, pd.DataFrame]) -> pd.Series:
    values = get_last_nonnan_values(df=df)
    if isinstance(df, pd.DataFrame):
        ds = pd.Series(values, index=df.columns)
    else:
        ds = pd.Series(values, index=[df.name])
    return ds


def multiply_df_by_dt(df: Union[pd.DataFrame, pd.Series],
                      dates: Union[pd.DatetimeIndex, pd.Index] = None,
                      lag: Optional[int] = None,
                      is_actual_calendar_dt: bool = True,
                      annualization_factor: float = 365.0
                      ) -> Union[pd.DataFrame, pd.Series]:
    """
    to compute rate adjustment with data - rate:
    get data at dates index and adjust by dt if needed
    adjust data by time spread:
    data = dt*data
    """
    if not isinstance(df.index, pd.DatetimeIndex):
        raise TypeError(f"data index must be DateTimeIndex: {df.index}")

    if lag is not None:
        if len(df.index) > lag:
            df = df.shift(lag)

    if dates is not None:
        # align tz
        if isinstance(dates, pd.DatetimeIndex) and df.index.tz is not None:
            df.index = df.index.tz_convert(dates.tz)
        df = df.reindex(index=dates, method='ffill')

    # apply dt multiplication
    if len(df.index) > 1:
        if is_actual_calendar_dt:
            delta = np.append(0.0, (df.index[1:] - df.index[:-1]).days / annualization_factor)
        else:
            delta = 1.0 / annualization_factor
        df = df.multiply(delta, axis=0)
    else:
        warnings.warn(f"in adjust_data_with_dt: lengh of data index is one - cannot adjust by dt")
        return df

    return df


def np_txy_tensor_to_pd_dict(np_tensor_txy: np.ndarray,
                             dateindex: Union[pd.Index, pd.DatetimeIndex],
                             factor_names: Union[pd.Index, List[str]],
                             asset_names: Union[pd.Index, List[str]],
                             is_factor_view: bool = True
                             ) -> Dict[str, pd.DataFrame]:
    """
    np_tensor_txy is an output of factor model: betas = [time, factor, asset]
    given timeindex such that #timeindex = time
    convert to:
    factor view = dict{factor: set[timeindex, asset]}  for each factor exp by assets
    asset view = dict{asset: set[timeindex, factor]}  for each asset exp by factor
    """
    if np_tensor_txy.ndim != 3:
        raise TypeError(f"np_tensor must ne 3-d tensor")
    if len(dateindex) != np_tensor_txy.shape[0]:
        raise TypeError(f"time dimension must be equal")
    if len(factor_names) != np_tensor_txy.shape[1]:
        raise TypeError(f"factor dimension must be equal")
    if len(asset_names) != np_tensor_txy.shape[2]:
        raise TypeError(f"asset dimension must be equal")

    if is_factor_view:  # {factor_id: pd.DataFrame(factor loadings)}
        factor_loadings = {}
        for factor_idx, factor in enumerate(factor_names):
            factor_loadings[factor] = pd.DataFrame(data=np_tensor_txy[:, factor_idx, :],
                                                   index=dateindex,
                                                   columns=asset_names)

    else:  # {asset_id: pd.DataFrame(factor loadings)}
        factor_loadings = {}
        for asset_idx, asset in enumerate(asset_names):
            factor_loadings[asset] = pd.DataFrame(data=np_tensor_txy[:, :, asset_idx],
                                                  index=dateindex,
                                                  columns=factor_names)
    return factor_loadings


def factor_dict_to_asset_dict(factor_loadings: Dict[str, pd.DataFrame],
                              asset_names: Optional[Union[pd.Index, List[str]]] = None  # column names in set
                              ) -> Dict[str, pd.DataFrame]:
    """
    revert the dictionary Dict[factor/model: set[assets] to Dict[assets: set factor)
    """
    if asset_names is None:
        asset_names = factor_loadings[list(factor_loadings.keys())[0]].columns

    asset_factor_loadings = {}
    for asset in asset_names:
        asset_factor_loading = []
        for factor, factor_data in factor_loadings.items():
            asset_factor_loading.append(factor_data[asset].rename(factor))
        asset_factor_loadings[asset] = pd.concat(asset_factor_loading, axis=1)

    return asset_factor_loadings


def df_time_dict_to_pd(factor_loadings_dict: Dict[pd.Timestamp, pd.DataFrame],
                       is_factor_view: bool = True,
                       is_zeros_to_nan: bool = False
                       ) -> Dict[str, pd.DataFrame]:
    """
    {time: set[factor, asset}
    convert to:
    factor view = dict{factor: set[timeindex, asset]}
    asset view = dict{asset: set[timeindex, factor]}
    """
    data0 = next(iter(factor_loadings_dict.values()))
    factor_names = data0.index
    asset_names = data0.columns
    dateindex = pd.DatetimeIndex(factor_loadings_dict.keys())

    np_tensor_txy = np.full((len(dateindex), len(factor_names), len(asset_names)), np.nan)
    for idx, data in enumerate(factor_loadings_dict.values()):
        np_tensor_txy[idx, :, :] = data.to_numpy()

    if is_zeros_to_nan:
        np_tensor_txy = np.where(np.isclose(np_tensor_txy, 0.0), np.nan, np_tensor_txy)

    factor_loadings = np_txy_tensor_to_pd_dict(np_tensor_txy=np_tensor_txy,
                                               dateindex=dateindex,
                                               factor_names=factor_names,
                                               asset_names=asset_names,
                                               is_factor_view=is_factor_view)
    return factor_loadings


def dfs_to_upper_lower_diag(df_upper: pd.DataFrame,
                            df_lower: pd.DataFrame,
                            diagonal: pd.Series
                            ) -> pd.DataFrame:
    """
    create up/low/diag copz
    nb must be square matrizes
    """
    nrows = len(df_upper.index)
    out = df_upper.copy()
    for i in range(nrows):
        out.iloc[i, i] = diagonal.iloc[i]
        for j in range(nrows):
            if i < j:
                out.iloc[i, j] = df_upper.iloc[i, j]
            elif i > j:
                out.iloc[i, j] = df_lower.iloc[i, j]
            else:
                pass
    return out


def compute_last_score(df: Union[pd.DataFrame, pd.Series], is_percent: bool = True) -> pd.Series:
    """
    columnwise score for last value in data
    use column wise loop with percentileofscore (as it supports only float for score)
    """
    if isinstance(df, pd.Series):
        df = df.to_frame()

    percentiles = []
    for column in df.columns:
        x = df[column].dropna()
        percentiles.append(stats.percentileofscore(a=x, score=x.iloc[-1], kind='rank'))
    percentiles = pd.Series(percentiles, index=df.columns)
    if is_percent:
        percentiles = percentiles.multiply(0.01)

    return percentiles


def align_dfs_dict_with_df(dfd: Dict[Any, pd.DataFrame],
                           df: pd.DataFrame,
                           fill_na_method: Optional[str] = 'ffill'
                           ) -> Dict[Any, pd.DataFrame]:
    """
    align dict of dataframes with given df
    """
    alignment_columns = df.columns
    alignment_index = df.index
    aligned_dfd = {}
    for key, df in dfd.items():
        if df is not None:  # align with other pandas in dict by filtered value
            df = df[alignment_columns].reindex(index=alignment_index)
            if fill_na_method is not None:
                if fill_na_method == 'ffill':
                    df = df.ffill()
                elif fill_na_method == 'bfill':
                    df = df.bfill()
                else:
                    raise NotImplementedError(f"fill_na_method={fill_na_method}")
            aligned_dfd[key] = df
        else:
            aligned_dfd[key] = None
    return aligned_dfd


def align_df1_to_df2(df1: pd.DataFrame,
                     df2: pd.DataFrame,
                     join: Literal['inner', 'outer', 'left', 'right'] = 'inner',
                     axis: Optional[int] = None
                     ) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    align dataframes
    join='inner' with strict match
    join='outer' without strict match so missing data has na
    """
    if isinstance(df1, pd.DataFrame) and isinstance(df2, pd.DataFrame):
        pass
    else:
        raise TypeError(f"df1 and df2 must be dataframes")

    df1_, df2_ = df1.align(other=df2, join=join, axis=axis)

    if join == 'inner':
        if len(df1.index) != len(df1_.index) or len(df2.index) != len(df2_.index):
            raise ValueError(f"df1 and df2 index are not matched: {df1.index} vs {df2.index}")

        elif len(df1.columns) != len(df1_.columns) or len(df2.columns) != len(df2_.columns):
            raise ValueError(f"df1 and df2 columns are not matched: {df1.columns} vs {df2.columns}")

    return df1_, df2_


def df12_merge_with_tz(df1: pd.DataFrame,
                       df2: pd.DataFrame,
                       tz: Optional[str] = None,
                       freq: str = 'B'
                       ) -> pd.DataFrame:
    """
    tz sensetive alignment
    """
    # localize index and convert to common time zeone
    df1 = df1.copy()
    df1.index = df1.index.tz_localize(tz=tz)
    df2 = df2.copy()
    df2.index = df2.index.tz_localize(tz=tz)

    # concat
    dfs = pd.concat([df1, df2], axis=1)
    dfs = dfs.ffill().asfreq(freq=freq, method='ffill')
    dfs.index = dfs.index# .tz_localize(tz=tz)
    return dfs


def merge_dfs_on_column(data_df: pd.DataFrame,
                        index_df: pd.DataFrame,
                        index_column_in_data_df: str = 'bbg_ticker'
                        ) -> pd.DataFrame:
    """
    merge data_df with index_df using index_column_in_data_df
    index_df
    """
    data_df.index.name = 'index1'  # rename index
    index_df.index.name = 'index2'  # rename index
    # align dfs by index_column_in_data_df
    index_df_joint_data = index_df.reindex(index=data_df[index_column_in_data_df].to_list())
    # merge aligned dfs with reset index
    joint_data = pd.concat([data_df.reset_index(drop=False), index_df_joint_data.reset_index(drop=False)], axis=1)
    joint_data = joint_data.set_index(data_df.index.name)
    return joint_data


def reindex_upto_last_nonnan(ds: pd.Series,
                             index: pd.DatetimeIndex,
                             method: str = 'ffill'
                             ) -> pd.Series:
    """
    apply ffill up to the last value
    """
    filled_ds = ds.reindex(index=index, method=method)
    last_non_nan = get_first_last_nonnan_index(df=ds, is_first=False)
    if filled_ds.index[-1] > last_non_nan:
        if last_non_nan in filled_ds.index:
            idx = filled_ds.index.get_loc(last_non_nan)  # find idx
            filled_ds.iloc[idx+1:] = np.nan
        else:
            filled_ds.loc[last_non_nan:] = np.nan
    return filled_ds


class UnitTests(Enum):
    ALIGN = 1
    SCORES = 2
    NONNANINDEX = 3
    REINDEX_UPTO_LAST_NONAN = 4
    MERGE_DFS_ON_COLUMNS = 5


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.ALIGN:

        desc_dict1 = {'f0': (0.5, 0.5), 'f1': (0.5, 0.5), 'f2': (0.5, 0.5), 'f3': (0.5, 0.5), 'f4': (0.5, 0.5)}
        df1 = pd.DataFrame.from_dict(desc_dict1, orient='index', columns=['as1', 'as2'])

        desc_dict2 = {'f1': (1.0, 1.0), 'f2': (1.0, 1.0), 'f5': (1.0, 1.0), 'f6': (1.0, 1.0)}
        df2 = pd.DataFrame.from_dict(desc_dict2, orient='index', columns=['as1', 'as3'])

        print(f"df1=\n{df1}")
        print(f"df2=\n{df2}")

        df1_, df2_ = align_df1_to_df2(df1=df1, df2=df2, join='outer', axis=0)
        print(f"df1_=\n{df1_}")
        print(f"df2_=\n{df2_}")

        df1_, df2_ = align_df1_to_df2(df1=df1, df2=df2, join='outer', axis=None)
        print(f"df1_=\n{df1_}")
        print(f"df2_=\n{df2_}")

    elif unit_test == UnitTests.SCORES:
        np.random.seed(1)
        nrows, ncols = 20, 5
        df = pd.DataFrame(data=np.random.normal(0.0, 1.0, size=(nrows, ncols)),
                          columns=[f"id{n + 1}" for n in range(ncols)])
        print(df)
        percentiles = compute_last_score(df=df)
        print(percentiles)

    elif unit_test == UnitTests.NONNANINDEX:

        values = [1.0, np.nan, 3.0, 4.0, np.nan, 6.0, np.nan, np.nan]
        dates = pd.date_range(start='1Jan2020', periods=len(values))
        df = pd.Series({d: v for d, v in zip(dates, values)})
        print(df)
        last_non_nan = get_first_last_nonnan_index(df=df, is_first=False)
        print(last_non_nan)

    elif unit_test == UnitTests.REINDEX_UPTO_LAST_NONAN:

        values = [1.0, np.nan, 3.0, 4.0, np.nan, 6.0, np.nan, 1.0]
        dates = pd.date_range(start='1Jan2020', periods=len(values))
        ds = pd.Series({d: v for d, v in zip(dates, values)})
        print(ds)

        dates1 = pd.date_range(start='1Jan2020', periods=len(values)+2)
        post_filled = ds.reindex(index=dates1, method='ffill')
        print(post_filled)

        post_filled_up_nan = reindex_upto_last_nonnan(ds=ds, index=dates1, method='ffill')
        print(post_filled_up_nan)

    elif unit_test == UnitTests.MERGE_DFS_ON_COLUMNS:
        data_entries = {'Bond1': pd.Series(['AAA', 100.00, 'A3'], index=['bbg_ticker', 'face', 'raiting']),
                        'Bond2': pd.Series(['AA', 100.00, 'A2'], index=['bbg_ticker', 'face', 'raiting']),
                        'Bond3': pd.Series(['A', 100.00, 'A1'], index=['bbg_ticker', 'face', 'raiting']),
                        'Bond4': pd.Series(['BBB', 100.00, 'B3'], index=['bbg_ticker', 'face', 'raiting'])}
        data_df = pd.DataFrame.from_dict(data_entries, orient='index')

        index_df = {'A': pd.Series([95.0], index=['price']),
                    'AA': pd.Series([99.0], index=['price']),
                    'AAA': pd.Series([101.0], index=['price']),
                    'B': pd.Series([90.0], index=['price'])}
        index_df = pd.DataFrame.from_dict(index_df, orient='index')
        print(data_df)
        print(index_df)
        df = merge_dfs_on_column(data_df=data_df, index_df=index_df)
        print(df)


if __name__ == '__main__':

    unit_test = UnitTests.MERGE_DFS_ON_COLUMNS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# imports
import matplotlib.pyplot as plt
import seaborn as sns
import yfinance as yf
import qis as qis
from qis import PerfStat

# define tickers and fetch price data
tickers = ['SPY', 'QQQ', 'EEM', 'TLT', 'IEF', 'SHY', 'LQD', 'HYG', 'GLD']
prices = yf.download(tickers, start=None, end=None)['Close'][tickers].dropna()

# define frequencies for vol computations
freqs = ['B', 'W-WED', 'ME', 'QE']

# plot scatter plot of PerfStat.SHARPE_RF0 vs PerfStat.SORTINO_RATIO
with sns.axes_style('darkgrid'):
    fig, axs = plt.subplots(len(freqs)//2, len(freqs)//2, figsize=(10, 5), tight_layout=True)
    qis.set_suptitle(fig, f"Sortino ratio vs Sharpe ratio with rf=0 as function of returns frequency")
    axs = qis.to_flat_list(axs)
    for idx, freq in enumerate(freqs):
        perf_table = qis.get_ra_perf_columns(prices=prices,
                                             perf_params=qis.PerfParams(freq=freq),
                                             perf_columns=[PerfStat.SHARPE_RF0, PerfStat.SORTINO_RATIO],
                                             is_to_str=False)
        qis.plot_scatter(df=perf_table,
                         x=PerfStat.SHARPE_RF0.to_str(),
                         y=PerfStat.SORTINO_RATIO.to_str(),
                         title=f"returns freq={freq}",
                         annotation_labels=perf_table.index.to_list(),
                         xvar_format='{:.2f}', yvar_format='{:.2f}',
                         full_sample_color='blue',
                         full_sample_order=1,
                         fit_intercept=False,
                         ax=axs[idx])

plt.show()



import matplotlib.pyplot as plt
import pandas as pd
import qis as qis
import qis.plots.utils as put
from typing import Optional, Tuple


def generate_price_history_report(prices: pd.DataFrame,
                                  figsize: Tuple[float, float] = (8.3, 11.7),
                                  **kwargs
                                  ) -> plt.Figure:
    fig, axs = plt.subplots(1, 2, figsize=figsize, tight_layout=True)
    qis.plot_ra_perf_table(prices=prices,
                           title='Risk-Adjusted Performance',
                           ax=axs[0],
                           **kwargs)
    plot_price_history(prices=prices,
                       title='Price History',
                       ax=axs[1],
                       **kwargs)
    return fig


def plot_price_history(prices: pd.DataFrame,
                       title: str = None,
                       #date_format: str = '%d-%b-%y',
                       ax: plt.Subplot = None,
                       **kwargs
                       ) -> Optional[plt.Figure]:

    start_dates = {}
    end_dates = {}
    durations = {}
    for asset in prices.columns:
        price = prices[asset].dropna()
        if not price.empty:
            start_dates[asset] = price.index[0]
            end_dates[asset] = price.index[-1]
            durations[asset] = qis.get_time_to_maturity(maturity_time=price.index[-1],
                                                        value_time=price.index[0])

    start_dates = pd.Series(start_dates).rename('Start')
    end_dates = pd.Series(end_dates).rename('End')
    durations = pd.Series(durations).rename('Period')
    df = pd.concat([start_dates, end_dates, durations], axis=1)
    df = df.iloc[::-1]  # reverse index
    if ax is None:
        width, height, _, _ = put.calc_df_table_size(df=df, min_rows=len(df.index), min_cols=len(df.index)//2)
        fig, ax = plt.subplots(1, 1, figsize=(width, height), constrained_layout=True)
    else:
        fig = None

    ax.hlines(df.index, xmin=df['Start'], xmax=df['End'])

    # put.set_ax_tick_params(ax=ax)
    put.set_ax_ticks_format(ax=ax, **qis.update_kwargs(kwargs, dict()))

    if title is not None:
        put.set_title(ax=ax, title=title, **kwargs)
    ax.margins(x=0.015, y=0.015)
    return fig



# imports
import matplotlib.pyplot as plt
import seaborn as sns
import yfinance as yf
import qis as qis


# define tickers and fetch price data
tickers = ['SPY', 'QQQ', 'EEM', 'TLT', 'IEF', 'SHY', 'LQD', 'HYG', 'GLD']
prices = yf.download(tickers, start=None, end=None)['Close'][tickers].dropna()

# minimum usage
with sns.axes_style("darkgrid"):
    fig, ax = plt.subplots(1, 1, figsize=(10, 7))
    qis.plot_prices(prices=prices, x_date_freq='YE', ax=ax)

# skip
qis.save_fig(fig, file_name='perf1', local_path="figures/")

# with drawdowns using sns styles
with sns.axes_style("darkgrid"):
    fig, axs = plt.subplots(2, 1, figsize=(10, 7))
    qis.plot_prices_with_dd(prices=prices, x_date_freq='YE', axs=axs)

# skip
qis.save_fig(fig, file_name='perf2', local_path="figures/")

# risk-adjusted performance table with specified data entries
# add rates for excess Sharpe
from qis import PerfStat
ust_3m_rate = yf.download('^IRX', start=None, end=None)['Close'].dropna() / 100.0

# set parameters for computing performance stats including returns vols and regressions
perf_params = qis.PerfParams(freq='ME', freq_reg='QE', alpha_an_factor=4.0, rates_data=ust_3m_rate)
# perf_columns is list to display different perfomance metrics from enumeration PerfStat
fig = qis.plot_ra_perf_table(prices=prices,
                             perf_columns=[PerfStat.TOTAL_RETURN, PerfStat.PA_RETURN, PerfStat.PA_EXCESS_RETURN,
                                           PerfStat.VOL, PerfStat.SHARPE_RF0,
                                           PerfStat.SHARPE_EXCESS, PerfStat.SORTINO_RATIO, PerfStat.CALMAR_RATIO,
                                           PerfStat.MAX_DD, PerfStat.MAX_DD_VOL,
                                           PerfStat.SKEWNESS, PerfStat.KURTOSIS],
                             title=f"Risk-adjusted performance: {qis.get_time_period_label(prices, date_separator='-')}",
                             perf_params=perf_params)

# skip
qis.save_fig(fig, file_name='perf3', local_path="figures/")

# add benchmark regression using excess returns for linear beta
# regression frequency is specified using perf_params.freq_reg
# regression alpha is multiplied using perf_params.alpha_an_factor
fig, _ = qis.plot_ra_perf_table_benchmark(prices=prices,
                                          benchmark='SPY',
                                          perf_columns=[PerfStat.TOTAL_RETURN, PerfStat.PA_RETURN, PerfStat.PA_EXCESS_RETURN,
                                                        PerfStat.VOL, PerfStat.SHARPE_RF0,
                                                        PerfStat.SHARPE_EXCESS, PerfStat.SORTINO_RATIO, PerfStat.CALMAR_RATIO,
                                                        PerfStat.MAX_DD, PerfStat.MAX_DD_VOL,
                                                        PerfStat.SKEWNESS, PerfStat.KURTOSIS,
                                                        PerfStat.ALPHA_AN, PerfStat.BETA, PerfStat.R2],
                                          title=f"Risk-adjusted performance: {qis.get_time_period_label(prices, date_separator='-')} benchmarked with SPY",
                                          perf_params=perf_params)
# skip
qis.save_fig(fig, file_name='perf4', local_path="figures/")

plt.show()



from __future__ import annotations

# packages
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from numba import njit
from dataclasses import dataclass, asdict
from statsmodels.regression.linear_model import RegressionResults as RegModel
from typing import Union, Dict, Any, Optional, Tuple, List
from enum import Enum

# qis
import qis as qis
from qis import (TimePeriod, PerfStat, PerfParams, RegimeData, EnumMap, BenchmarkReturnsQuantileRegimeSpecs,
                 RollingPerfStat)
import qis.utils.df_groups as dfg
import qis.utils.df_agg as dfa
import qis.perfstats.returns as ret
import qis.perfstats.perf_stats as rpt
import qis.perfstats.regime_classifier as rcl
import qis.plots.time_series as pts
import qis.plots.stackplot as pst
import qis.plots.derived.prices as ppd
import qis.plots.derived.perf_table as ppt
import qis.plots.derived.returns_scatter as prs
import qis.plots.derived.returns_heatmap as rhe
import qis.models.linear.ewm_factors as ef
from qis.models.linear.ewm import compute_ewm_vol
from qis.portfolio.ewm_portfolio_risk import compute_portfolio_vol, compute_portfolio_risk_contributions


# default performance and regime params
PERF_PARAMS = PerfParams(freq='W-WED')
REGIME_PARAMS = BenchmarkReturnsQuantileRegimeSpecs(freq='ME')


class AttributionMetric(str, Enum):
    """
    input for computation of get_performance_attribution_data()
    """
    PNL = 'P&L Attribution, sum=portfolio performance'
    PNL_RISK = 'P&L Risk Attribution, sum=100%'
    INST_PNL = 'Instrument P&L'
    COSTS = 'Instrument Total Costs'
    TURNOVER = 'Instrument Annualised Turnover'
    VOL_ADJUSTED_TURNOVER = 'Instrument Annualised Vol-Adjusted Turnover'


class SnapshotPeriod(str, Enum):
    """
    input for computation of get_performance_attribution_data()
    """
    LAST = 'last'
    AVG = 'avg'
    MAX = 'max'


@dataclass
class PortfolioData:
    """
    portfolio data can be generated by
    1. qis.backtest_model_portfolio()
    2. independent backtester with outputs matching PortfolioData inputs
    it is possible to supply StrategySignalData for 2
    """
    nav: pd.Series  # nav is computed with all cost
    weights: pd.DataFrame = None  # weights of portfolio
    units: pd.DataFrame = None  # units of portfolio instruments
    prices: pd.DataFrame = None  # prices of portfolio universe
    instrument_pnl: pd.DataFrame = None  # include net pnl by instrument
    realized_costs: pd.DataFrame = None  # realized trading costs by instrument
    input_weights: Union[np.ndarray, pd.DataFrame, Dict[str, float]] = None  # inputs to potfolio
    is_rebalancing: pd.Series = None  # optional info when the weights are rebalanced
    tickers_to_names_map: Optional[Dict[str, str]] = None  # renaming of long tickers
    group_data: pd.Series = None  # for asset class grouping
    group_order: List[str] = None
    instrument_names: pd.Series = None
    benchmark_prices: pd.DataFrame = None  # can pass benchmark prices here
    ticker: str = None
    strategy_signal_data: StrategySignalData = None
    covar_dict: Dict[pd.Timestamp, pd.DataFrame] = None  # for computing risk contributions

    def __post_init__(self):
        if isinstance(self.nav, pd.DataFrame):
            self.nav = self.nav.iloc[:, 0]
        if self.prices is None:
            self.prices = self.nav.to_frame()
        if self.weights is None:  # default will be delta-1 portfolio of nav
            self.weights = pd.DataFrame(1.0, index=self.prices.index, columns=self.prices.columns)
        if self.units is None:  # default will be delta-1 portfolio of nav
            self.units = pd.DataFrame(1.0, index=self.prices.index, columns=self.prices.columns)
        if self.realized_costs is None:
            self.realized_costs = pd.DataFrame(0.0, index=self.prices.index, columns=self.prices.columns)
        if self.instrument_pnl is None:
            self.instrument_pnl = self.prices.pct_change(fill_method=None).multiply(self.weights.shift(1)).fillna(0.0)
        if self.group_data is None:  # use instruments as groups
            self.group_data = pd.Series(self.prices.columns, index=self.prices.columns)
        if self.group_order is None:
            self.group_order = list(self.group_data.unique())
        if self.benchmark_prices is not None:
            self.ibenchmark_prices = self.benchmark_prices.reindex(index=self.nav.index, method='ffill')
        if self.ticker is None:
            self.ticker = str(self.nav.name)

    def set_ticker(self, ticker: str) -> PortfolioData:
        self.ticker = ticker
        self.nav = self.nav.rename(ticker)
        return self

    def set_benchmark_prices(self, benchmark_prices: Union[pd.Series, pd.DataFrame]) -> None:
        # can pass benchmark prices here
        if isinstance(benchmark_prices, pd.Series):
            benchmark_prices = benchmark_prices.to_frame()
        self.benchmark_prices = benchmark_prices.reindex(index=self.nav.index, method='ffill')

    def set_group_data(self, group_data: pd.Series, group_order: List[str] = None) -> None:
        self.group_data = group_data
        if group_order is None:
            group_order = group_data.to_list()
        self.group_order = group_order

    def save(self, ticker: str, local_path: str = './') -> None:
        datasets = dict(nav=self.nav, prices=self.prices, weights=self.weights, units=self.units,
                        instrument_pnl=self.instrument_pnl, realized_costs=self.realized_costs)
        if self.group_data is not None:
            datasets['group_data'] = self.group_data
        qis.save_df_dict_to_csv(datasets=datasets, file_name=ticker, local_path=local_path)
        print(f"saved portfolio data for {ticker}")

    @classmethod
    def load(cls, ticker: str) -> PortfolioData:
        dataset_keys = ['nav', 'prices', 'weights', 'units', 'instrument_pnl', 'realized_costs', 'group_data']
        datasets = qis.load_df_dict_from_csv(dataset_keys=dataset_keys, file_name=ticker)
        return cls(**datasets)

    """
    NAV level getters
    """

    def get_portfolio_nav(self, time_period: TimePeriod = None, freq: Optional[str] = None,
                          ticker: Optional[str] = None) -> pd.Series:
        """
        get nav using consistent function for all return computations
        """
        if time_period is not None:
            nav_ = time_period.locate(self.nav)
        else:
            nav_ = self.nav.copy()
        if freq is not None:
            nav_ = nav_.asfreq(freq=freq, method='ffill')
        if ticker is not None:
            nav_ = nav_.rename(ticker)
        return nav_

    def get_portfolio_nav_with_benchmark_prices(self,
                                               time_period: TimePeriod = None,
                                               freq: Optional[str] = None
                                               ) -> pd.DataFrame:
        """
        get nav using consistent function for all return computations
        """
        navs = self.get_portfolio_nav(time_period=time_period, freq=freq)
        if self.benchmark_prices is not None:
            benchmark_prices = self.benchmark_prices.reindex(index=navs.index, method='ffill')
            navs = pd.concat([navs, benchmark_prices], axis=1)
        return navs

    def get_instruments_pnl(self,
                            add_total: bool = False,
                            time_period: TimePeriod = None,
                            is_net: bool = False,
                            is_unit_based_traded_volume: bool = True,
                            is_compounded: bool = False,
                            freq: Optional[str] = None
                            ) -> pd.DataFrame:
        pnl = self.instrument_pnl.copy()
        if is_net:
            costs = self.get_costs(add_total=False, is_unit_based_traded_volume=is_unit_based_traded_volume)
            pnl = pnl.subtract(costs)
        if add_total:
            pnl.insert(loc=0, value=pnl.sum(1), column='Total')
        if time_period is not None:
            pnl = time_period.locate(pnl)
        if freq is not None:
            pnl = pnl.resample(freq).sum()
        if is_compounded:
            pnl = np.expm1(pnl)
        return pnl

    def get_performance_attribution(self,
                                    add_total: bool = True,
                                    time_period: TimePeriod = None,
                                    is_compounded: bool = False
                                    ) -> pd.Series:
        instrument_pnl = self.get_instruments_pnl(add_total=add_total,
                                                  time_period=time_period,
                                                  is_compounded=is_compounded)
        if is_compounded:
            performance_attribution = instrument_pnl.iloc[-1, :] - 1.0
        else:
            performance_attribution = instrument_pnl.sum(axis=0)
        return performance_attribution

    def get_instruments_navs(self,
                             time_period: TimePeriod = None,
                             constant_trade_level: bool = False
                             ) -> pd.DataFrame:
        pnl = self.get_instruments_pnl(time_period=time_period, is_compounded=False).fillna(0.0)
        navs = ret.returns_to_nav(returns=pnl, constant_trade_level=constant_trade_level)
        return navs

    def get_group_navs(self,
                       time_period: TimePeriod = None,
                       constant_trade_level: bool = False,
                       is_add_group_total: bool = False
                       ) -> Optional[pd.DataFrame]:
        """
        group total will exclude transaction costs so it is not equal to portfolio nav
        """
        if is_add_group_total:
            total_column = str(self.nav.name)
        else:
            total_column = None
        df = self.get_instruments_pnl(time_period=time_period)
        if df.empty:
            print(f"instruments p&l is not available for time_period={time_period.to_str()}")
            group_navs = None
        else:
            grouped_pnl = dfg.agg_df_by_groups_ax1(df=df,
                                                   group_data=self.group_data,
                                                   agg_func=np.nansum,
                                                   total_column=total_column,
                                                   group_order=self.group_order)
            group_navs = ret.returns_to_nav(returns=grouped_pnl, constant_trade_level=constant_trade_level)
        return group_navs

    def get_total_nav_with_group_navs(self, time_period: TimePeriod = None) -> pd.DataFrame:
        """
        group total will exclude transaction costs so it is not equal to portfolio nav
        """
        total_nav = self.get_portfolio_nav(time_period=time_period)
        group_navs = self.get_group_navs(time_period=time_period, is_add_group_total=False)
        prices = pd.concat([total_nav, group_navs], axis=1)
        return prices

    def get_input_weights(self, time_period: TimePeriod = None) -> Union[np.ndarray, pd.DataFrame, Dict[str, float]]:
        input_weights = self.input_weights.copy()
        if time_period is not None and isinstance(input_weights, pd.DataFrame):
            input_weights = time_period.locate(input_weights)
        return input_weights

    def get_weights(self,
                    is_input_weights: bool = False,
                    columns: List[str] = None,
                    time_period: TimePeriod = None,
                    is_grouped: bool = False,
                    add_total: bool = False,
                    freq: Optional[str] = 'W-WED',
                    group_data: pd.Series = None,
                    group_order: List[str] = None
                    ) -> pd.DataFrame:
        if is_input_weights and self.input_weights is not None and isinstance(self.input_weights, pd.DataFrame):
            weights = self.input_weights.copy()
        else:
            weights = self.weights.copy()
        if time_period is not None:
            weights = time_period.locate(weights)
        if freq is not None:
            weights = weights.resample(freq).last().ffill()

        if is_grouped:
            if group_data is None:
                group_data = self.group_data
            if group_order is None:
                group_order = self.group_order
            weights = dfg.agg_df_by_groups_ax1(df=weights,
                                               group_data=group_data,
                                               agg_func=np.nansum,
                                               total_column=self.ticker if add_total else None,
                                               group_order=group_order)
        elif columns is not None:
            weights = weights[columns]

        return weights

    def get_grouped_long_short_exposures(self,
                                         time_period: TimePeriod = None,
                                         total_name: str = 'Total',
                                         exposures_freq: Optional[str] = 'W-WED'
                                         ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, pd.DataFrame]]:
        """
        compute grouped net long / short exposures
        """
        group_dict = dfg.get_group_dict(group_data=self.group_data,
                                        group_order=self.group_order,
                                        total_column=None)
        all_exposures = self.get_weights(time_period=time_period, freq=exposures_freq)
        grouped_exposures_by_inst = {}
        grouped_exposures_agg = {}
        for group, tickers in group_dict.items():
            exposures_by_inst = all_exposures[tickers]
            grouped_exposures_by_inst[group] = exposures_by_inst
            total = dfa.nansum(exposures_by_inst, axis=1).rename(total_name)
            net_long = dfa.nansum_positive(exposures_by_inst, axis=1).rename('Net Long')
            net_short = dfa.nansum_negative(exposures_by_inst, axis=1).rename('Net Short')
            grouped_exposures_agg[group] = pd.concat([total, net_long, net_short], axis=1)
        return grouped_exposures_agg, grouped_exposures_by_inst

    def get_grouped_cum_pnls(self,
                             total_name: str = 'Total',
                             time_period: TimePeriod = None
                             ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, pd.DataFrame]]:
        """
        compute grouped net long / short exposues
        """
        group_dict = dfg.get_group_dict(group_data=self.group_data,
                                        group_order=self.group_order,
                                        total_column=None)

        pnls = self.get_instruments_pnl(time_period=time_period, is_compounded=False).fillna(0.0)
        all_exposures = self.get_weights(time_period=time_period, freq=None)
        all_exposures = all_exposures.reindex(index=pnls.index, method='ffill').ffill()
        pnl_positive_exp = pnls.where(all_exposures > 0.0, other=0.0)
        pnl_negative_exp = pnls.where(all_exposures < 0.0, other=0.0)

        grouped_pnls_by_inst = {}
        grouped_pnls_agg = {}
        for group, tickers in group_dict.items():
            pnls_by_inst = pnls[tickers]
            grouped_pnls_by_inst[group] = pnls_by_inst.cumsum(axis=0)
            total = dfa.nansum(pnls_by_inst, axis=1).rename(total_name)
            net_long = dfa.nansum(pnl_positive_exp[tickers], axis=1).rename('Net Long')
            net_short = dfa.nansum(pnl_negative_exp[tickers], axis=1).rename('Net Short')
            grouped_pnls_agg[group] = pd.concat([total, net_long, net_short], axis=1).cumsum(axis=0)
        return grouped_pnls_agg, grouped_pnls_by_inst

    def get_turnover(self,
                     is_agg: bool = False,
                     is_grouped: bool = False,
                     group_data: pd.Series = None,
                     group_order: List[str] = None,
                     time_period: TimePeriod = None,
                     roll_period: Optional[int] = 260,
                     is_vol_adjusted: bool = False,
                     add_total: bool = True,
                     vol_span: int = 33,
                     freq: Optional[str] = None,
                     is_unit_based_traded_volume: bool = True
                     ) -> Union[pd.DataFrame, pd.Series]:

        if is_unit_based_traded_volume:  # for unit generated backtest
            turnover = (self.units.diff(1).abs()).multiply(self.prices)
            abs_exposure = self.units.multiply(self.prices).abs().sum(axis=1)
            # turnover = turnover.divide(self.nav.to_numpy(), axis=0)
            turnover = turnover.divide(abs_exposure.to_numpy(), axis=0)
        else:  # for weight generated backtets
            turnover = self.input_weights.diff(1).abs()

        if is_vol_adjusted:
            instrument_vols = compute_ewm_vol(data=qis.to_returns(self.prices, is_log_returns=True),
                                              span=vol_span,
                                              annualize=True)
            turnover = turnover.multiply(instrument_vols)

        if is_agg:
            turnover = pd.Series(np.nansum(turnover, axis=1), index=turnover.index, name=self.nav.name)
            turnover = turnover.reindex(index=self.nav.index)
        elif is_grouped:  # agg by groups
            if group_data is None:
                group_data = self.group_data
            if group_order is None:
                group_order = self.group_order
            turnover = dfg.agg_df_by_groups_ax1(df=turnover,
                                                group_data=group_data,
                                                agg_func=np.nansum,
                                                total_column=str(self.nav.name) if add_total else None,
                                                group_order=group_order)
        else:
            if add_total:
                turnover = pd.concat([turnover.sum(axis=1).rename(self.nav.name), turnover], axis=1)

        if freq is not None:  # first aggregate by freq
            turnover = turnover.resample(freq).sum()
        if roll_period is not None:  # now aggregate by roll
            turnover = turnover.rolling(roll_period).sum()
        if time_period is not None:
            turnover = time_period.locate(turnover)
        return turnover

    def get_costs(self,
                  is_agg: bool = False,
                  is_grouped: bool = False,
                  time_period: TimePeriod = None,
                  add_total: bool = True,
                  is_unit_based_traded_volume: bool = True,
                  roll_period: Optional[int] = 260,
                  freq: Optional[str] = None
                  ) -> Union[pd.DataFrame, pd.Series]:

        costs = self.realized_costs
        if is_unit_based_traded_volume:
            costs = costs.divide(self.nav.to_numpy(), axis=0)
        if is_agg:
            costs = pd.Series(np.nansum(costs, axis=1), index=self.nav.index, name=self.nav.name)
        elif is_grouped:  # agg by groups
            costs = dfg.agg_df_by_groups_ax1(costs,
                                             group_data=self.group_data,
                                             agg_func=np.nansum,
                                             total_column=str(self.nav.name) if add_total else None,
                                             group_order=self.group_order)
        else:
            if add_total:
                costs = pd.concat([costs.sum(axis=1).rename(self.nav.name), costs], axis=1)

        if freq is not None:  # first aggregate by freq
            costs = costs.resample(freq).sum()
        if roll_period is not None:  # now aggregate by roll period
            costs = costs.rolling(roll_period).sum()
        if time_period is not None:
            costs = time_period.locate(costs)
        return costs

    def compute_mcap_participation(self,
                                   mcap: pd.DataFrame,
                                   trade_level: float = 100000000
                                   ) -> pd.DataFrame:
        exposure = (self.units.multiply(self.prices)).divide(self.nav.to_numpy(), axis=0)
        participation = trade_level * exposure.divide(mcap)
        return participation

    def compute_volume_participation(self,
                                     volumes: pd.DataFrame,
                                     trade_level: float = 100000000
                                     ) -> pd.DataFrame:
        turnover = self.get_turnover(is_agg=False, is_grouped=False)
        participation = trade_level * turnover.divide(volumes)
        return participation

    def compute_cumulative_attribution(self) -> pd.DataFrame:
        attribution = (self.prices.pct_change()).multiply(self.weights.shift(1))
        attribution = attribution.cumsum(axis=0)
        return attribution

    def compute_realized_pnl(self, time_period: TimePeriod = None) -> Tuple[pd.DataFrame, ...]:
        avg_costs, realized_pnl, mtm_pnl, trades = compute_realized_pnl(prices=self.prices.to_numpy(),
                                                                        units=self.units.to_numpy())
        avg_costs = pd.DataFrame(avg_costs, index=self.prices.index, columns=self.prices.columns)
        realized_pnl = pd.DataFrame(realized_pnl, index=self.prices.index, columns=self.prices.columns)
        mtm_pnl = pd.DataFrame(mtm_pnl, index=self.prices.index, columns=self.prices.columns)
        trades = pd.DataFrame(trades, index=self.prices.index, columns=self.prices.columns)
        if time_period is not None:
            avg_costs = time_period.locate(avg_costs)
            realized_pnl = time_period.locate(realized_pnl)
            mtm_pnl = time_period.locate(mtm_pnl)
            trades = time_period.locate(trades)
        realized_pnl = realized_pnl.cumsum(axis=0)
        total_pnl = realized_pnl.add(mtm_pnl)
        return avg_costs, realized_pnl, mtm_pnl, total_pnl, trades

    def compute_portfolio_benchmark_betas(self,
                                          benchmark_prices: pd.DataFrame,
                                          time_period: TimePeriod = None,
                                          freq_beta: str = None,
                                          factor_beta_span: int = 65  # quarter
                                          ) -> pd.DataFrame:
        """
        compute benchmark betas of instruments
        portfolio_beta_i = sum(instrument_beta_i*exposure)
        """
        instrument_prices = self.prices
        benchmark_prices = benchmark_prices.reindex(index=instrument_prices.index, method='ffill')
        exposures = self.get_weights().reindex(index=instrument_prices.index, method='ffill')
        benchmark_betas = ef.compute_portfolio_benchmark_betas(instrument_prices=instrument_prices,
                                                               exposures=exposures,
                                                               benchmark_prices=benchmark_prices,
                                                               time_period=time_period,
                                                               freq_beta=freq_beta,
                                                               factor_beta_span=factor_beta_span)
        return benchmark_betas

    def compute_portfolio_benchmark_attribution(self,
                                                benchmark_prices: pd.DataFrame,
                                                time_period: TimePeriod = None,
                                                freq_beta: str = 'B',
                                                factor_beta_span: int = 63,  # quarter
                                                residual_name: str = 'Alpha'
                                                ) -> pd.DataFrame:
        """
        attribution = portfolio_return_{t} - benchmark_return_{t}*bet_{t-1}
        returns are compounded
        """
        instrument_prices = self.prices
        benchmark_prices = benchmark_prices.reindex(index=instrument_prices.index, method='ffill')
        exposures = self.get_weights().reindex(index=instrument_prices.index, method='ffill')
        portfolio_nav = self.get_portfolio_nav().reindex(index=instrument_prices.index, method='ffill')
        joint_attrib = ef.compute_portfolio_benchmark_beta_alpha_attribution(instrument_prices=instrument_prices,
                                                                             exposures=exposures,
                                                                             benchmark_prices=benchmark_prices,
                                                                             portfolio_nav=portfolio_nav,
                                                                             time_period=time_period,
                                                                             freq_beta=freq_beta,
                                                                             factor_beta_span=factor_beta_span,
                                                                             residual_name=residual_name)
        return joint_attrib

    """
    ### instrument level getters
    """

    def get_instruments_returns(self,
                                time_period: TimePeriod = None
                                ) -> pd.DataFrame:
        returns = self.prices.pct_change(fill_method=None)
        if time_period is not None:
            returns = time_period.locate(returns)
        return returns

    def get_instruments_periodic_returns(self,
                                         time_period: TimePeriod = None,
                                         freq: str = 'ME'
                                         ) -> pd.DataFrame:
        returns = self.get_instruments_returns(time_period=time_period)
        prices = ret.returns_to_nav(returns=returns, init_period=None)
        returns_f = ret.to_returns(prices=prices, freq=freq)
        return returns_f

    def get_instruments_performance_attribution(self,
                                                time_period: TimePeriod = None,
                                                constant_trade_level: bool = False
                                                ) -> pd.Series:
        navs = self.get_instruments_navs(time_period=time_period, constant_trade_level=constant_trade_level)
        perf = ret.to_total_returns(prices=navs).rename(self.nav.name)
        return perf

    def get_instruments_pnl_risk_attribution(self,
                                             time_period: TimePeriod = None
                                             ) -> pd.DataFrame:
        pnl = self.get_instruments_pnl(time_period=time_period)
        # portfolio_pnl = pnl.sum(axis=1)

        pnl_risk = np.nanstd(pnl.replace({0.0: np.nan}), axis=0)
        # portfolio_pnl_risk = np.nanstd(portfolio_pnl.replace({0.0: np.nan}), axis=0)
        # pnl_risk_ratio = pnl_risk / portfolio_pnl_risk
        pnl_risk_ratio = pnl_risk / np.nansum(pnl_risk)

        data = pd.Series(pnl_risk_ratio, index=pnl.columns, name=self.nav.name)
        if self.tickers_to_names_map is not None:
            data = data.rename(index=self.tickers_to_names_map)
        return data

    def get_performance_attribution_data(self,
                                         attribution_metric: AttributionMetric = AttributionMetric.PNL,
                                         time_period: TimePeriod = None,
                                         is_unit_based_traded_volume: bool = True,
                                         **kwargs
                                         ) -> Union[pd.DataFrame, pd.Series]:
        if attribution_metric == AttributionMetric.PNL:
            data = self.get_instruments_performance_attribution(time_period=time_period)
        elif attribution_metric == AttributionMetric.PNL_RISK:
            data = self.get_instruments_pnl_risk_attribution(time_period=time_period)
        elif attribution_metric == AttributionMetric.INST_PNL:
            data = self.get_instruments_navs(time_period=time_period)
        elif attribution_metric == AttributionMetric.COSTS:
            data = self.get_costs(is_agg=False, is_grouped=False, roll_period=None, is_unit_based_traded_volume=is_unit_based_traded_volume,
                                  add_total=False,
                                  time_period=time_period)
            data = data.sum(0)
            #print(f"total costs, is_unit_based_traded_volume={is_unit_based_traded_volume} = {np.nansum(data)}")
        elif attribution_metric == AttributionMetric.TURNOVER:
            data = self.get_turnover(is_agg=False, is_grouped=False, roll_period=None,
                                     add_total=False,
                                     time_period=time_period)
            an = qis.infer_an_from_data(data=data)
            data = an*data.mean(0)
            #print(f"total turnover = {np.nansum(data)}")
        elif attribution_metric == AttributionMetric.VOL_ADJUSTED_TURNOVER:
            data = self.get_turnover(is_agg=False, is_grouped=False, roll_period=None,
                                     add_total=False, is_vol_adjusted=True,
                                     time_period=time_period)
            an = qis.infer_an_from_data(data=data)
            data = an * data.mean(0)

        else:
            raise NotImplementedError(f"{attribution_metric}")

        return data

    def get_num_investable_instruments(self, time_period: TimePeriod = None) -> pd.DataFrame:
        num_available_prices = np.sum(np.where(np.isfinite(self.prices), 1.0, 0.0), axis=1)
        num_available_prices = pd.Series(num_available_prices, index=self.weights.index, name='Investable')
        non_zero_exposures = self.weights.where(np.isclose(self.weights, 0.0) == False, other=np.nan)
        num_invested_instruments = np.sum(np.where(np.isfinite(non_zero_exposures), 1.0, 0.0), axis=1)
        num_invested_instruments = pd.Series(num_invested_instruments, index=self.weights.index, name='Invested')
        df = pd.concat([num_available_prices, num_invested_instruments], axis=1)
        if time_period is not None:
            df = time_period.locate(df)
        return df

    def get_instruments_performance_table(self,
                                          time_period: TimePeriod = None,
                                          portfolio_name: str = 'Attribution'
                                          ) -> pd.DataFrame:
        """
        using avg weight
        """
        insts_returns = self.get_instruments_returns(time_period=time_period)
        insts_return = ret.to_total_returns(prices=ret.returns_to_nav(returns=insts_returns))
        weight = self.weights
        if time_period is not None:
            weight = time_period.locate(weight)
        weight = weight.mean(axis=0)
        portf_return = insts_return.multiply(weight).replace({0.0: np.nan}).dropna()
        data = pd.concat([weight.rename('Weight'),
                          insts_return.rename('Asset'),
                          portf_return.rename(portfolio_name)],
                         axis=1).dropna()
        data = data.sort_values('Weight', ascending=False)
        if self.tickers_to_names_map is not None:
            data = data.rename(index=self.tickers_to_names_map)

        return data

    def get_attribution_table_by_instrument(self,
                                            time_period: TimePeriod = None,
                                            freq: str = 'ME',
                                            is_input_weights: bool = False
                                            ) -> pd.DataFrame:
        """
        using avg weight
        """
        if is_input_weights:
            weights = self.input_weights
            if not isinstance(weights, pd.DataFrame):
                raise ValueError(f"input weights must be pd.Dataframe for is_input_weights=True")
            prices_w = self.prices.reindex(index=weights.index, method='ffill')
            returns_f = prices_w.pct_change()
            portf_return = returns_f.multiply(weights.shift(1)).iloc[1:, :]
        else:
            returns_f = self.get_instruments_periodic_returns(time_period=time_period, freq=freq)
            weight = self.weights.reindex(index=returns_f.index, method='ffill')
            # first row is None
            portf_return = returns_f.multiply(weight.shift(1)).iloc[1:, :]
        if self.tickers_to_names_map is not None:
            portf_return = portf_return.rename(columns=self.tickers_to_names_map)
        return portf_return

    def compute_portfolio_vol(self,
                              time_period: TimePeriod = None,
                              freq: str = 'W-WED',
                              span: int = 13  # 3m span of weekly returns
                              ) -> pd.DataFrame:
        """
        compute_portfolio_vol using 1) instrument weights and covar matrix 2) realised returns vol
        """
        returns_f = self.get_instruments_periodic_returns(freq=freq)
        weights = self.weights.reindex(index=returns_f.index, method='ffill')
        portfolio_vol = compute_portfolio_vol(returns=returns_f,
                                              weights=weights,
                                              span=span,
                                              annualize=True)
        strategy_vol = compute_ewm_vol(data=qis.to_returns(self.get_portfolio_nav(freq=freq), is_log_returns=True),
                                       span=span,
                                       annualize=True)
        df = pd.concat([portfolio_vol.rename('instrument weighted vol'),
                        strategy_vol.rename('strategy returns vol')
                        ], axis=1)
        if time_period is not None:
            df = time_period.locate(df)
        return df

    def compute_distribution_yield(self,
                                   paid_dividends: pd.DataFrame,
                                   div_rolling_freq: str = 'ME',
                                   div_rolling_period: int = 12,
                                   time_period: TimePeriod = None
                                   ) -> Tuple[pd.DataFrame, pd.Series, pd.Series]:
        positions = self.units
        distributions_by_instrument = paid_dividends.multiply(positions)
        nav = self.get_portfolio_nav().reindex(index=distributions_by_instrument.index, method='ffill')
        distributions_by_instrument_yield = distributions_by_instrument.divide(nav, axis=0)
        # resample at 'ME' and compute rolling sum
        distributions_by_instrument_yield_freq = distributions_by_instrument_yield.resample(div_rolling_freq).sum()
        distributions_by_instrument_yield_rolling = distributions_by_instrument_yield_freq.fillna(0.0).rolling(div_rolling_period).sum()
        # find total distributions
        distribution_yield = distributions_by_instrument_yield_freq.sum(1)
        distribution_yield_rolling = distributions_by_instrument_yield_rolling.sum(1)
        if time_period is not None:
            distributions_by_instrument_yield_rolling = time_period.locate(distributions_by_instrument_yield_rolling)
            distribution_yield = time_period.locate(distribution_yield)
            distribution_yield_rolling = time_period.locate(distribution_yield_rolling)
        return distributions_by_instrument_yield_rolling, distribution_yield, distribution_yield_rolling

    def compute_portfolio_vars(self,
                               is_correlated: bool = True,
                               time_period: TimePeriod = None,
                               freq: str = 'B',
                               total_column: Optional[str] = 'Total',
                               vol_span: Union[int, float] = 33,  # span in number of freq-returns
                               group_data: pd.Series = None,
                               group_order: List[str] = None
                               ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        if group_data is None:
            group_data = self.group_data
        if group_order is None:
            group_order = self.group_order
        if is_correlated:
            portfolio_vars = qis.compute_portfolio_correlated_var_by_groups(prices=self.prices,
                                                                            weights=self.get_weights(freq=freq),
                                                                            group_data=group_data,
                                                                            group_order=group_order,
                                                                            freq=freq,
                                                                            vol_span=vol_span,  # span in number of freq-returns
                                                                            total_column=total_column,
                                                                            time_period=time_period)
            instrument_vars = None
        else:
            instrument_vars, portfolio_vars = qis.compute_portfolio_independent_var_by_ac(prices=self.prices,
                                                                                          weights=self.get_weights(freq=freq),
                                                                                          group_data=group_data,
                                                                                          group_order=group_order,
                                                                                          freq=freq,
                                                                                          vol_span=vol_span,  # span in number of freq-returns
                                                                                          total_column=total_column,
                                                                                          time_period=time_period)
        return portfolio_vars, instrument_vars

    def compute_risk_contributions_implied_by_covar(self,
                                                    covar_dict: Dict[pd.Timestamp, pd.DataFrame] = None,
                                                    group_data: pd.Series = None,
                                                    group_order: List[str] = None,
                                                    align_with_covar_dates: bool = True,
                                                    freq: Optional[str] = None,
                                                    normalise: bool = False
                                                    ) -> pd.DataFrame:
        """
        compute risk contributions using covar_dict
        """
        if covar_dict is None:
            if self.covar_dict is None:
                raise ValueError(f"must provide covar_dict")
            else:
                covar_dict = self.covar_dict
        strategy_weights = self.get_weights(freq=freq, is_input_weights=True)
        covar_index = list(covar_dict.keys())
        strategy_rc = {}
        if align_with_covar_dates:
            strategy_weights = strategy_weights.reindex(index=covar_index).ffill().fillna(0.0)
            for date, pd_covar in covar_dict.items():
                strategy_rc[date] = compute_portfolio_risk_contributions(w=strategy_weights.loc[date], covar=pd_covar)
        else:
            for date, weights in strategy_weights.to_dict(orient='index').items():
                last_covar_update_date = qis.find_upto_date_from_datetime_index(index=covar_index, date=date)
                if last_covar_update_date is not None:
                    strategy_rc[date] = compute_portfolio_risk_contributions(w=pd.Series(weights).fillna(0.0),
                                                                             covar=covar_dict[last_covar_update_date])

        strategy_rc = pd.DataFrame.from_dict(strategy_rc, orient='index')
        if group_data is not None:
            strategy_rc = dfg.agg_df_by_groups_ax1(strategy_rc, group_data=group_data, group_order=group_order)
        if normalise:
            strategy_rc = qis.df_to_weight_allocation_sum1(strategy_rc)
        return strategy_rc

    # """
    # plotting methods
    # """

    def add_regime_shadows(self,
                           ax: plt.Subplot,
                           regime_benchmark: str,
                           index: pd.Index = None,
                           regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS
                           ) -> None:
        """
        add regime shadows using regime_benchmark
        """
        if self.benchmark_prices is None:
            raise ValueError(f"set benchmarks data")
        pivot_prices = self.benchmark_prices[regime_benchmark]
        if index is not None:
            pivot_prices = pivot_prices.reindex(index=index, method='ffill')
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)

    def plot_nav(self,
                 regime_benchmark: str = None,
                 time_period: TimePeriod = None,
                 add_benchmarks: bool = False,
                 regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                 ax: plt.Subplot = None,
                 **kwargs
                 ) -> None:
        if add_benchmarks:
            prices = self.get_portfolio_nav_with_benchmark_prices(time_period=time_period)
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
        if ax is None:
            with sns.axes_style('darkgrid'):
                fig, ax = plt.subplots(1, 1, figsize=(16, 12), tight_layout=True)
        ppd.plot_prices(prices=prices, ax=ax, **kwargs)

        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=prices.index,
                                    regime_params=regime_params)

    def plot_group_nav(self,
                       regime_benchmark: str = None,
                       time_period: TimePeriod = None,
                       add_benchmarks: bool = False,
                       regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                       ax: plt.Subplot = None,
                       **kwargs
                       ) -> None:

        total_group_navs = self.get_total_nav_with_group_navs(time_period=time_period)
        if add_benchmarks and self.benchmark_prices is not None:
            benchmark_prices = self.benchmark_prices.reindex(index=total_group_navs.index, method='ffill')
            total_group_navs = pd.concat([total_group_navs, benchmark_prices], axis=1)
        if ax is None:
            with sns.axes_style('darkgrid'):
                fig, ax = plt.subplots(1, 1, figsize=(16, 12), tight_layout=True)
        ppd.plot_prices(prices=total_group_navs, ax=ax, **kwargs)

        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=total_group_navs.index,
                                    regime_params=regime_params)

    def plot_portfolio_vols(self,
                            time_period: TimePeriod = None,
                            freq: str = 'W-WED',
                            span: int = 13,  # 3m span of weekly returns
                            regime_benchmark: str = None,
                            regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                            ax: plt.Subplot = None,
                            **kwargs
                            ) -> plt.Figure:
        if ax is None:
            fig, ax = plt.subplots()

        portfolio_vols = self.compute_portfolio_vol(time_period=time_period, freq=freq, span=span)
        fig = qis.plot_time_series(df=portfolio_vols,
                                   var_format='{:,.2%}',
                                   legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                                   title=f"Portfolio EWM {span}-span vols of {freq}-freq returns",
                                   ax=ax,
                                   **kwargs)
        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=portfolio_vols.index,
                                    regime_params=regime_params)
        return fig

    def plot_rolling_perf(self,
                          rolling_perf_stat: RollingPerfStat = RollingPerfStat.SHARPE,
                          add_benchmarks: bool = False,
                          regime_benchmark: str = None,
                          time_period: TimePeriod = None,
                          rolling_window: int = 1300,
                          roll_freq: Optional[str] = None,
                          legend_stats: pts.LegendStats = pts.LegendStats.AVG_LAST,
                          title: Optional[str] = None,
                          regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> plt.Figure:

        # do not use start end dates here so the sharpe will be continuous with different time_period
        if add_benchmarks:
            prices = self.get_portfolio_nav_with_benchmark_prices(time_period=time_period)
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
        if ax is None:
            fig, ax = plt.subplots()
        fig = ppd.plot_rolling_perf_stat(prices=prices,
                                         rolling_perf_stat=rolling_perf_stat,
                                         time_period=time_period,
                                         roll_periods=rolling_window,
                                         roll_freq=roll_freq,
                                         legend_stats=legend_stats,
                                         title=title,
                                         ax=ax,
                                         **kwargs)
        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=prices.index,
                                    regime_params=regime_params)
        return fig

    def plot_ra_perf_table(self,
                           benchmark_price: Union[pd.Series, pd.DataFrame] = None,
                           benchmark: str = None,
                           is_grouped: bool = True,
                           time_period: TimePeriod = None,
                           perf_params: PerfParams = None,
                           perf_columns: List[PerfStat] = rpt.BENCHMARK_TABLE_COLUMNS,
                           title: str = None,
                           ax: plt.Subplot = None,
                           **kwargs
                           ) -> None:
        if is_grouped:
            total_nav = self.get_portfolio_nav(time_period=time_period)
            group_navs = self.get_group_navs(time_period=time_period, is_add_group_total=False)
            prices = pd.concat([total_nav, group_navs], axis=1)
            for_title = 'with portfolio groups'
            rows_edge_lines = [len(prices.columns)]
        else:
            prices = self.get_portfolio_nav(time_period=time_period).to_frame()
            for_title = ''
            rows_edge_lines = None
        if benchmark_price is not None:
            if isinstance(benchmark_price, pd.DataFrame):  # bechmark is the last asset
                if benchmark is None:
                    benchmark = benchmark_price.columns[-1]
            else:
                benchmark = benchmark_price.name
            if benchmark not in prices.columns:
                prices = pd.concat([prices, benchmark_price.reindex(index=prices.index, method='ffill')], axis=1)
            prices = prices.loc[:, ~prices.columns.duplicated(keep='first')]
            title = title or f"RA performance table {for_title} for {perf_params.freq_vol}-freq returns with beta to {benchmark}:" \
                             f" {qis.get_time_period(prices).to_str()}"
            ppt.plot_ra_perf_table_benchmark(prices=prices,
                                             benchmark=benchmark,
                                             perf_params=perf_params,
                                             perf_columns=perf_columns,
                                             title=title,
                                             rotation_for_columns_headers=0,
                                             special_rows_colors=[(1, 'deepskyblue'),
                                                                  (len(prices.columns), 'lightskyblue')],
                                             rows_edge_lines=rows_edge_lines,
                                             column_header='Portfolio',
                                             ax=ax,
                                             **kwargs)
        else:
            title = title or f"RA performance table {for_title}: {qis.get_time_period(prices).to_str()}"
            ppt.plot_ra_perf_table(prices=prices,
                                   perf_params=perf_params,
                                   perf_columns=rpt.COMPACT_TABLE_COLUMNS,
                                   title=title,
                                   rotation_for_columns_headers=0,
                                   column_header='Portfolio',
                                   ax=ax,
                                   **kwargs)

    def plot_returns_scatter(self,
                             benchmark_price: pd.Series,
                             is_grouped: bool = True,
                             time_period: TimePeriod = None,
                             title: str = None,
                             freq: str = 'QE',
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        if is_grouped:
            prices = self.get_total_nav_with_group_navs(time_period=time_period)
            title = title or f"Scatterplot of {freq}-freq returns by groups vs {str(benchmark_price.name)}"
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
            title = title or f"Scatterplot of {freq}-freq returns vs {str(benchmark_price.name)}"
        prices = pd.concat([prices, benchmark_price.reindex(index=prices.index, method='ffill')], axis=1)
        if np.any(prices.columns.duplicated()):
            raise ValueError(f"duplicated columns {prices.columns}")
        local_kwargs = qis.update_kwargs(kwargs=kwargs,
                                         new_kwargs={'weight': 'bold',
                                                     'x_rotation': 0,
                                                     'first_color_fixed': False,
                                                     'ci': None,
                                                     'markersize': 6})
        prs.plot_returns_scatter(prices=prices,
                                 benchmark=str(benchmark_price.name),
                                 freq=freq,
                                 order=2,
                                 title=title,
                                 ax=ax,
                                 **local_kwargs)

    def plot_monthly_returns_heatmap(self,
                                     time_period: TimePeriod = None,
                                     ax: plt.Subplot = None,
                                     **kwargs
                                     ) -> None:
        # for monthly returns fix A and date_format
        kwargs = qis.update_kwargs(kwargs, dict(heatmap_freq='YE', date_format='%Y'))
        rhe.plot_returns_heatmap(prices=self.get_portfolio_nav(time_period=time_period),
                                 heatmap_column_freq='ME',
                                 is_add_annual_column=True,
                                 is_inverse_order=True,
                                 ax=ax,
                                 **kwargs)

    def plot_periodic_returns(self,
                              benchmark_prices: Union[pd.DataFrame, pd.Series] = None,
                              is_grouped: bool = True,
                              time_period: TimePeriod = None,
                              heatmap_freq: str = 'YE',
                              date_format: str = '%Y',
                              transpose: bool = True,
                              title: str = None,
                              ax: plt.Subplot = None,
                              **kwargs
                              ) -> None:
        if is_grouped:
            prices = self.get_total_nav_with_group_navs(time_period=time_period)
            title = title or f"{heatmap_freq}-returns by groups"
        else:
            prices = self.get_portfolio_nav(time_period=time_period).to_frame()
            title = title or f"{heatmap_freq}-returns"
        if benchmark_prices is not None:
            hline_rows = [1, len(prices.columns)]
            prices = pd.concat([prices, benchmark_prices.reindex(index=prices.index, method='ffill')], axis=1)
        else:
            hline_rows = None

        rhe.plot_periodic_returns_table(prices=prices,
                                        ax=ax,
                                        title=title,
                                        date_format=date_format,
                                        transpose=transpose,
                                        hline_rows=hline_rows,
                                        **qis.update_kwargs(kwargs, dict(freq=heatmap_freq)))

    def plot_regime_data(self,
                         benchmark_price: pd.Series,
                         is_grouped: bool = True,
                         regime_data_to_plot: RegimeData = RegimeData.REGIME_SHARPE,
                         time_period: TimePeriod = None,
                         var_format: Optional[str] = None,
                         is_conditional_sharpe: bool = True,
                         legend_loc: Optional[str] = 'upper center',
                         title: str = None,
                         perf_params: PerfParams = None,
                         regime_params: BenchmarkReturnsQuantileRegimeSpecs = None,
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:
        regime_classifier = rcl.BenchmarkReturnsQuantilesRegime(regime_params=regime_params)

        if is_grouped:
            prices = self.get_total_nav_with_group_navs(time_period=time_period)
            title = title or (f"Sharpe ratio attribution by groups to {str(benchmark_price.name)} "
                              f"Bear/Normal/Bull regimes of {regime_classifier.regime_params.freq} returns")
        else:
            prices = self.get_portfolio_nav(time_period=time_period).to_frame()
            title = title or (f"Sharpe ratio attribution to {str(benchmark_price.name)} Bear/Normal/Bull regimes "
                              f"of {regime_classifier.regime_params.freq} returns")

        if benchmark_price.name not in prices.columns:
            prices = pd.concat([benchmark_price.reindex(index=prices.index, method='ffill'), prices], axis=1)

        fig = qis.plot_regime_data(regime_classifier=regime_classifier,
                                   prices=prices,
                                   benchmark=str(benchmark_price.name),
                                   is_conditional_sharpe=is_conditional_sharpe,
                                   regime_data_to_plot=regime_data_to_plot,
                                   var_format=var_format or '{:.2f}',
                                   legend_loc=legend_loc,
                                   perf_params=perf_params,
                                   title=title,
                                   ax=ax,
                                   **kwargs)
        return fig

    def plot_vol_regimes(self,
                         benchmark_price: pd.Series,
                         is_grouped: bool = True,
                         time_period: TimePeriod = None,
                         title: str = None,
                         freq: str = 'ME',
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:

        if is_grouped:
            prices = self.get_total_nav_with_group_navs(time_period=time_period)
            title = title or f"{freq}-returns by groups conditional on vols {str(benchmark_price.name)}"
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
            title = title or f"{freq}-returns conditional on vols {str(benchmark_price.name)}"
        prices = pd.concat([benchmark_price.reindex(index=prices.index, method='ffill'), prices], axis=1)

        regime_classifier = rcl.BenchmarkVolsQuantilesRegime(regime_params=rcl.VolQuantileRegimeSpecs(freq=freq))
        fig = qis.plot_regime_boxplot(regime_classifier=regime_classifier,
                                      prices=prices,
                                      benchmark=str(benchmark_price.name),
                                      title=title,
                                      ax=ax,
                                      **kwargs)
        return fig

    def plot_contributors(self,
                          time_period: TimePeriod = None,
                          num_assets: int = 10,
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> None:
        prices = self.get_instruments_navs(time_period=time_period)
        ppt.plot_top_bottom_performers(prices=prices, num_assets=num_assets, ax=ax, **kwargs)

    def plot_best_worst_returns(self,
                                time_period: TimePeriod = None,
                                num_returns: int = 10,
                                ax: plt.Subplot = None,
                                **kwargs
                                ) -> None:
        price = self.get_portfolio_nav(time_period=time_period)
        ppt.plot_best_worst_returns(price=price, num_returns=num_returns, ax=ax, **kwargs)

    def plot_pnl(self, time_period: TimePeriod = None) -> None:
        avg_costs, realized_pnl, mtm_pnl, total_pnl, trades = self.compute_realized_pnl(time_period=time_period)
        prices = self.prices
        if time_period is not None:
            prices = time_period.locate(prices)
        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(5, 1, figsize=(10, 16), tight_layout=True)
            pts.plot_time_series(df=prices, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='prices', ax=axs[0])
            pts.plot_time_series(df=avg_costs, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='avg_costs',
                                 ax=axs[1])
            pts.plot_time_series(df=realized_pnl, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='realized_pnl',
                                 ax=axs[2])
            pts.plot_time_series(df=mtm_pnl, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='mtm_pnl', ax=axs[3])
            pts.plot_time_series(df=total_pnl, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='total_pnl',
                                 ax=axs[4])

    def plot_weights(self,
                     is_input_weights: bool = True,
                     add_mean_levels: bool = False,
                     use_bar_plot: bool = False,
                     columns: List[str] = None,
                     freq: Optional[str] = None,
                     is_yaxis_limit_01: bool = True,
                     bbox_to_anchor: Tuple[float, float] = None,
                     legend_stats: pst.LegendStats = pst.LegendStats.FIRST_AVG_LAST,
                     var_format: str = '{:.1%}',
                     title: Optional[str] = None,
                     ax: plt.Subplot = None,
                     **kwargs
                     ) -> None:
        weights = self.get_weights(is_input_weights=is_input_weights,
                                   columns=columns,
                                   freq=freq)
        pst.plot_stack(df=weights,
                       add_mean_levels=add_mean_levels,
                       use_bar_plot=use_bar_plot,
                       is_yaxis_limit_01=is_yaxis_limit_01,
                       bbox_to_anchor=bbox_to_anchor,
                       title=title,
                       legend_stats=legend_stats,
                       var_format=var_format,
                       ax=ax,
                       **kwargs)

    def plot_performance_attribution(self,
                                     time_period: TimePeriod = None,
                                     attribution_metric: AttributionMetric = AttributionMetric.PNL,
                                     add_top_bar_values: Optional[bool] = None,
                                     remove_zero_data: bool = False,
                                     ax: plt.Subplot = None,
                                     **kwargs
                                     ) -> None:
        """
        performance attribution for p&l and risk
        """
        data = self.get_performance_attribution_data(attribution_metric=attribution_metric,
                                                     time_period=time_period,
                                                     **kwargs)
        if remove_zero_data:
            data = data.replace({0.0: np.nan}).dropna()

        if add_top_bar_values is None:
            if isinstance(data, pd.Series) and len(data.index) <= 20:
                add_top_bar_values = True
            else:
                add_top_bar_values = False

        kwargs = qis.update_kwargs(kwargs=kwargs,
                                   new_kwargs=dict(bbox_to_anchor=(0.5, 1.05),
                                                   add_top_bar_values=add_top_bar_values,
                                                   x_rotation=90))
        title = f"{attribution_metric.value}"
        if time_period is not None:
            title += f", for period={time_period.to_str()}"
        qis.plot_bars(df=data,
                      skip_y_axis=True,
                      title=title,
                      stacked=False,
                      yvar_format='{:,.1%}' if add_top_bar_values else '{:,.2%}',
                      ax=ax,
                      **kwargs)

    def plot_benchmark_betas(self,
                             benchmark_prices: pd.DataFrame,
                             regime_benchmark: str = None,
                             time_period: TimePeriod = None,
                             freq: str = 'W-WED',
                             title: str = None,
                             beta_span: int = 52,
                             regime_params: BenchmarkReturnsQuantileRegimeSpecs = None,
                             add_zero_line: bool = True,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        factor_exposures = self.compute_portfolio_benchmark_betas(benchmark_prices=benchmark_prices,
                                                                  time_period=time_period,
                                                                  freq_beta=freq,
                                                                  factor_beta_span=beta_span)
        qis.plot_time_series(df=factor_exposures,
                             var_format='{:,.2f}',
                             legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                             title=title or f"Portfolio rolling {beta_span}-span Betas to Benchmarks",
                             ax=ax,
                             **kwargs)
        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=factor_exposures.index,
                                    regime_params=regime_params)

        if add_zero_line:
            ax.axhline(0, color='black', lw=1)

    def plot_portfolio_grouped_var(self,
                                   is_correlated: bool = True,
                                   regime_benchmark: str = None,
                                   time_period: TimePeriod = None,
                                   freq: str = 'B',
                                   title: str = None,
                                   total_column: Optional[str] = 'Total',
                                   vol_span: int = 33,  # span in number of freq-returns
                                   regime_params: BenchmarkReturnsQuantileRegimeSpecs = None,
                                   ax: plt.Subplot = None,
                                   **kwargs
                                   ) -> None:
        portfolio_vars, instrument_vars = self.compute_portfolio_vars(is_correlated=is_correlated,
                                                                      time_period=time_period,
                                                                      freq=freq,
                                                                      total_column=total_column,
                                                                      vol_span=vol_span)
        if is_correlated:
            title = title or f"Correlated {freq}-freq 99%-VAR with {vol_span}-span ewma covar"
        else:
            title = title or f"Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols"

        qis.plot_time_series(df=portfolio_vars,
                             var_format='{:,.2%}',
                             legend_stats=qis.LegendStats.AVG_MIN_MAX_LAST,
                             title=title,
                             y_limits=(0.0, None),
                             ax=ax,
                             **kwargs)

        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=portfolio_vars.index,
                                    regime_params=regime_params)

    def plot_current_weights(self,
                             is_grouped: bool = False,
                             add_top_bar_values: Optional[bool] = None,
                             time_period: TimePeriod = None,
                             title: str = None,
                             group_data: pd.Series = None,
                             group_order: List[str] = None,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        weights = self.get_weights(is_input_weights=True, freq=None, is_grouped=is_grouped,
                                   group_data=group_data, group_order=group_order)
        if time_period is not None:
            weights_1 = time_period.locate(weights)
            if len(weights_1.index) > 0:
                weights_1 = weights_1.iloc[0, :]
            else:
                weights_1 = weights.iloc[-1, :]
        else:
            weights_1 = weights.iloc[-1, :]
        if add_top_bar_values is None:
            if len(weights_1.index) <= 10:
                add_top_bar_values = True
            else:
                add_top_bar_values = False
        kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None,
                                                                  add_top_bar_values=add_top_bar_values,
                                                                  x_rotation=90))
        if title is None:
            if is_grouped:
                title = f"Weights by groups on {qis.date_to_str(weights_1.name)}"
            else:
                title = f"Weights by instruments on {qis.date_to_str(weights_1.name)}"
        qis.plot_bars(df=weights_1,
                      skip_y_axis=True,
                      title=title,
                      stacked=False,
                      yvar_format='{:,.2%}' if add_top_bar_values else '{:,.2%}',
                      ax=ax,
                      **kwargs)

    def plot_last_weights_change(self,
                                 is_grouped: bool = False,
                                 title: str = None,
                                 add_top_bar_values: Optional[bool] = None,
                                 weights_change_lag: Optional[int] = None,
                                 ax: plt.Subplot = None,
                                 **kwargs
                                 ) -> None:
        weights = self.get_weights(is_input_weights=True, freq=None, is_grouped=is_grouped)
        weights_1 = weights.iloc[-1, :]
        if len(weights.index) > 1:
            if weights_change_lag is not None:
                weights_0 = weights.iloc[-weights_change_lag, :]
            else:
                weights_0 = weights.iloc[-2, :]
            delta = weights_1.subtract(weights_0)
            post_title = f"rebalancing between {qis.date_to_str(weights_0.name)} and {qis.date_to_str(weights_1.name)}"
        else:
            delta = weights_1
            post_title = f"rebalancing on {qis.date_to_str(weights_1.name)} starting from zero"
        if add_top_bar_values is None:
            if len(delta.index) <= 10:
                add_top_bar_values = True
            else:
                add_top_bar_values = False
        kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None,
                                                                  add_top_bar_values=add_top_bar_values,
                                                                  x_rotation=90))
        if title is None:
            if is_grouped:
                title = f"Change in weights by groups with {post_title}"
            else:
                title = f"Change in weights by instruments with {post_title}"
        qis.plot_bars(df=delta,
                      skip_y_axis=True,
                      title=title,
                      stacked=False,
                      yvar_format='{:,.2%}' if add_top_bar_values else '{:,.2%}',
                      ax=ax,
                      **kwargs)

    def plot_current_var(self,
                         snapshot_period: SnapshotPeriod = SnapshotPeriod.LAST,
                         is_grouped: bool = False,
                         is_correlated: bool = True,
                         time_period: TimePeriod = None,
                         freq: str = 'B',
                         title: str = None,
                         add_top_bar_values: Optional[bool] = None,
                         total_column: Optional[str] = 'Total',
                         vol_span: Union[int, float] = 33,  # span in number of freq-returns
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> None:
        portfolio_vars, instrument_vars = self.compute_portfolio_vars(is_correlated=is_correlated,
                                                                      time_period=time_period,
                                                                      freq=freq,
                                                                      total_column=total_column,
                                                                      vol_span=vol_span)
        if is_grouped:
            if snapshot_period == SnapshotPeriod.LAST:
                var_1 = portfolio_vars.iloc[-1, :]
                if is_correlated:
                    title = title or f"Correlated {freq}-freq 99%-VAR with {vol_span}-span ewma covar: {qis.date_to_str(var_1.name)}"
                else:
                    title = title or f"Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols: {qis.date_to_str(var_1.name)}"
            elif snapshot_period == SnapshotPeriod.AVG:
                var_1 = portfolio_vars.mean(0)
                period = qis.get_time_period(df=portfolio_vars).to_str(date_separator='-')
                if is_correlated:
                    title = title or f"Avg Correlated {freq}-freq 99%-VAR with {vol_span}-span: {period}"
                else:
                    title = title or f"Avg Independent {freq}-freq 99%-VAR with {vol_span}-span: {period}"
            elif snapshot_period == SnapshotPeriod.MAX:
                var_1 = portfolio_vars.max(0)
                period = qis.get_time_period(df=portfolio_vars).to_str(date_separator='-')
                if is_correlated:
                    title = title or f"Max Correlated {freq}-freq 99%-VAR with {vol_span}-span: {period}"
                else:
                    title = title or f"Max Independent {freq}-freq 99%-VAR with {vol_span}-span: {period}"
            else:
                raise NotImplementedError(f"snapshot_period={snapshot_period}")

        else:
            if is_correlated:
                raise ValueError(f"instrument var is not defined for correlated var ")
            else:
                if snapshot_period == SnapshotPeriod.LAST:
                    var_1 = instrument_vars.iloc[-1, :]
                    title = title or f"Instrument independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols: {qis.date_to_str(var_1.name)}"
                elif snapshot_period == SnapshotPeriod.AVG:
                    var_1 = instrument_vars.mean(0)
                    period = qis.get_time_period(df=instrument_vars).to_str()
                    title = title or f"Avg Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols: {period}"
                elif snapshot_period == SnapshotPeriod.MAX:
                    var_1 = instrument_vars.max(0)
                    period = qis.get_time_period(df=instrument_vars).to_str()
                    title = title or f"Max Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols: {period}"
                else:
                    raise NotImplementedError(f"snapshot_period={snapshot_period}")

        if add_top_bar_values is None:
            if len(var_1.index) <= 10:
                add_top_bar_values = True
            else:
                add_top_bar_values = False
        kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None,
                                                                  add_bar_values=add_top_bar_values,
                                                                  x_rotation=90))
        qis.plot_bars(df=var_1,
                      skip_y_axis=True,
                      title=title,
                      stacked=False,
                      yvar_format='{:,.2%}',
                      ax=ax,
                      **kwargs)

    def plot_var_stack(self,
                       is_grouped: bool = True,
                       is_correlated: bool = True,
                       time_period: TimePeriod = None,
                       freq: str = 'B',
                       title: str = None,
                       stack_freq: Optional[str] = 'W-WED',
                       vol_span: int = 33,  # span in number of freq-returns
                       ax: plt.Subplot = None,
                       **kwargs
                       ) -> None:
        portfolio_vars, instrument_vars = self.compute_portfolio_vars(is_correlated=is_correlated,
                                                                      time_period=time_period,
                                                                      freq=freq,
                                                                      total_column=None,
                                                                      vol_span=vol_span)
        if is_grouped:
            var_1 = portfolio_vars
            if is_correlated:
                title = title or f"Risk Attribution with Correlated {freq}-freq 99%-VAR with {vol_span}-span ewma covar"
            else:
                title = title or f"Risk Attribution with Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols "

        else:
            if is_correlated:
                raise ValueError(f"instrument var is not defined for correlated var ")
            else:
                var_1 = instrument_vars
                title = title or f"Risk Attribution with Instrument independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols"

        if time_period is not None:
            var_1 = time_period.locate(var_1)
        if stack_freq is not None:
            var_1 = var_1.asfreq(stack_freq, method='ffill')
        df = var_1.divide(np.nansum(var_1, axis=1, keepdims=True))
        qis.plot_stack(df=df,
                       is_yaxis_limit_01=True,
                       use_bar_plot=True,
                       title=title,
                       legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                       var_format='{:.1%}',
                       ax=ax,
                       **qis.update_kwargs(kwargs, dict(bbox_to_anchor=(0.5, 1.05), ncols=2)))

    def plot_distribution_yield(self,
                                paid_dividends: pd.DataFrame,
                                time_period: TimePeriod = None,
                                div_rolling_freq: str = 'ME',
                                div_rolling_period: int = 12,
                                axs: List[plt.Subplot] = None,
                                **kwargs
                                ):
        distributions_by_instrument_yield_12m, distribution_yield, distribution_yield_12m =\
            self.compute_distribution_yield(paid_dividends=paid_dividends,
                                            time_period=time_period,
                                            div_rolling_freq=div_rolling_freq,
                                            div_rolling_period=div_rolling_period)
        if axs is None:
            fig1, ax1 = plt.subplots(1, 1, figsize=(14, 10), constrained_layout=True)
            fig2, ax2 = plt.subplots(1, 1, figsize=(14, 10), constrained_layout=True)
            axs = [ax1, ax2]

        qis.plot_time_series(distribution_yield_12m,
                             ax=axs[0],
                             **kwargs)
        qis.plot_stack(df=distributions_by_instrument_yield_12m.resample('ME').last(),
                       use_bar_plot=True,
                       ax=axs[1],
                       **kwargs)


@dataclass
class StrategySignalData:
    """
    data class instance applied for output of strategy backtest data
    """
    log_returns: pd.DataFrame = None
    ra_carry: pd.DataFrame = None  # risk-adjusted carry
    momentum: pd.DataFrame = None
    signal: pd.DataFrame = None  # signal output
    instrument_vols: pd.DataFrame = None  # instrument vols
    instrument_target_vols: pd.DataFrame = None  # target vols for portfolio allocation
    instrument_target_signal_vol_weights: pd.DataFrame = None  # target vols * signal
    instrument_portfolio_leverages: pd.DataFrame = None  # = portfolio weight / instrument_target_signal_vol_weights
    weights: pd.DataFrame = None  # final weights
    kwargs: Dict[str, pd.DataFrame] = None  # any other outputs

    def locate_period(self, time_period: TimePeriod) -> StrategySignalData:
        # nb: does not work for returns
        data_dict = asdict(self)
        for key, df in data_dict.items():
            if df is not None:
                data_dict[key] = time_period.locate(df)
        return StrategySignalData(**data_dict)

    def rename_data(self, names_map: Dict[str, str]) -> StrategySignalData:
        data_dict = asdict(self)
        for key, df in data_dict.items():
            if df is not None:
                data_dict[key] = df.rename(names_map, axis=1)
        return StrategySignalData(**data_dict)

    def get_current_signal_by_groups(self, group_data: pd.Series,
                                     group_order: List[str] = None
                                     ) -> Dict[str, pd.DataFrame]:
        group_dict = dfg.get_group_dict(group_data=group_data,
                                        group_order=group_order,
                                        total_column=None)
        group_signals = {}
        agg_by_group = {}
        last_date = qis.date_to_str(self.signal.index[-21])
        current_date = qis.date_to_str(self.signal.index[-1])
        last_signals = self.signal.iloc[-21, :]
        current_signals = self.signal.iloc[-1, :]
        for group, tickers in group_dict.items():
            last_signals_ = last_signals[tickers]
            current_signals_ = current_signals[tickers]
            group_signals[group] = pd.concat([last_signals_.rename(last_date),
                                              current_signals_.rename(current_date)
                                              ], axis=1)

            agg_by_group[group] = pd.Series({last_date: np.nanmean(last_signals_),
                                             current_date: np.nanmean(current_signals_)})
        agg_by_group = {'Total by groups': pd.DataFrame.from_dict(agg_by_group, orient='index')}
        agg_by_group.update(group_signals)
        return agg_by_group

    def asdiff(self, tickers: List[str] = None,
               freq: str = None,
               sample_size: Optional[int] = 21,  # can use rolling instead for freq
               time_period: TimePeriod = None
               ) -> StrategySignalData:

        if tickers is None:
            tickers = self.log_returns.columns.to_list()
        # nb: does not work for returns
        if time_period is not None:
            ssd = self.locate_period(time_period=time_period)
        else:
            ssd = self

        data_dict = asdict(ssd)
        if sample_size is not None:
            for key, df in data_dict.items():
                if df is not None:
                    data_dict[key] = qis.df_resample_at_int_index(df=df[tickers], func=None, sample_size=sample_size).diff()
        else:
            for key, df in data_dict.items():
                if df is not None:
                    data_dict[key] = qis.df_resample_at_freq(df=df[tickers], freq=freq, include_end_date=True).diff()
        return StrategySignalData(**data_dict)

    def estimate_signal_changes_joint(self,
                                      tickers: List[str] = None,
                                      freq: Optional[str] = None,
                                      sample_size: Optional[int] = 21,  # can use rolling instead for freq
                                      time_period: TimePeriod = None
                                      ) -> Tuple[pd.DataFrame, RegModel, TimePeriod]:
        if tickers is None:
            tickers = self.log_returns.columns.to_list()
        ssd = self.asdiff(tickers=tickers, sample_size=sample_size, freq=freq, time_period=time_period)
        y_var_name = 'weight_change'
        y = qis.melt_df_by_columns(ssd.weights.iloc[:-1, :], y_var_name=y_var_name)[y_var_name]
        x_var_name1 = 'momentum_change'
        x1 = qis.melt_df_by_columns(ssd.momentum.iloc[:-1, :], y_var_name=x_var_name1)[x_var_name1]
        x_var_name2 = 'target_vol_change'
        x2 = qis.melt_df_by_columns(ssd.instrument_target_vols.iloc[:-1, :], y_var_name=x_var_name2)[x_var_name2]
        x_var_name3 = 'port_leverage_change'
        x3 = qis.melt_df_by_columns(ssd.instrument_portfolio_leverages.iloc[:-1, :], y_var_name=x_var_name3)[x_var_name3]
        if self.ra_carry is not None:
            x_var_name0 = 'carry_change'
            x0 = qis.melt_df_by_columns(ssd.ra_carry.iloc[:-1, :], y_var_name=x_var_name0)[x_var_name0]
            x = pd.concat([x0, x1, x2, x3], axis=1).dropna()
        else:
            x = pd.concat([x1, x2, x3], axis=1).dropna()
        x_names = x.columns.to_list()
        y = y.reindex(index=x.index)

        # keep last obs for prediction
        fitted_model = qis.fit_ols(x=x.to_numpy(), y=y.to_numpy(), order=1, fit_intercept=False)
        actual_change = ssd.weights.iloc[-1, :]
        predictions = {}
        for ticker in tickers:
            if self.ra_carry is not None:
                x_ts = np.array([ssd.ra_carry[ticker].iloc[-1],
                                 ssd.momentum[ticker].iloc[-1],
                                 ssd.instrument_target_vols[ticker].iloc[-1],
                                 ssd.instrument_portfolio_leverages[ticker].iloc[-1]])
            else:
                x_ts = np.array([ssd.momentum[ticker].iloc[-1],
                                 ssd.instrument_target_vols[ticker].iloc[-1],
                                 ssd.instrument_portfolio_leverages[ticker].iloc[-1]])
            pred_t = {}
            total_pred = 0.0
            for idx, x_t in enumerate(x_ts):
                pred_x = fitted_model.params[idx] * x_t
                total_pred += pred_x
                pred_t[x_names[idx]] = pred_x
            pred_t['predicted'] = total_pred
            pred_t['actual'] = actual_change[ticker]
            pred_t['residual'] = actual_change[ticker] - total_pred
            pred_t['residual %'] = total_pred / actual_change[ticker]
            pred_t['r2'] = fitted_model.rsquared
            predictions[ticker] = pd.Series(pred_t)

        predictions = pd.DataFrame.from_dict(predictions, orient='index')
        prediction_period = TimePeriod(start=ssd.weights.index[-2], end=ssd.weights.index[-1])

        return predictions, fitted_model, prediction_period

    def estimate_signal_changes_by_groups(self,
                                          group_data: pd.Series, group_order: List[str] = None,
                                          freq: Optional[str] = None,
                                          sample_size: Optional[int] = 21,  # can use rolling instead for freq
                                          time_period: TimePeriod = None
                                          ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, RegModel], TimePeriod]:
        """
        estimate weight change for groups
        """
        group_dict = dfg.get_group_dict(group_data=group_data,
                                        group_order=group_order,
                                        total_column=None)
        predictions = {}
        fitted_models = {}
        prediction_period = None
        for group, tickers in group_dict.items():
            prediction, fitted_model, prediction_period = self.estimate_signal_changes_joint(
                tickers=tickers, freq=freq,
                sample_size=sample_size,
                time_period=time_period)
            predictions[group] = prediction
            fitted_models[group] = fitted_model
        return predictions, fitted_models, prediction_period

    def estimate_signal_changes_individual(self,
                                           tickers: List[str] = None,
                                           freq: Optional[str] = None,
                                           sample_size: Optional[int] = 21,  # can use rolling instead for freq
                                           time_period: TimePeriod = None
                                           ) -> Tuple[pd.DataFrame, Dict[str, RegModel]]:
        if tickers is None:
            tickers = self.log_returns.columns.to_list()
        ssd = self.asdiff(tickers=tickers, sample_size=sample_size, freq=freq, time_period=time_period)
        y_var_name = 'weight_change'
        y = ssd.weights
        x_var_name1 = 'momentum_change'
        x1 = ssd.momentum
        x_var_name2 = 'target_vol_change'
        x2 = ssd.instrument_target_vols
        x_var_name3 = 'port_leverage_change'
        x3 = ssd.instrument_portfolio_leverages
        if self.ra_carry is not None:
            x_var_name0 = 'carry_change'
            x0 = ssd.ra_carry
            x_names = [x_var_name0, x_var_name1, x_var_name2, x_var_name3]
        else:
            x_names = [x_var_name1, x_var_name2, x_var_name3]

        predictions = {}
        fitted_models = {}
        for ticker in tickers:
            if self.ra_carry is not None:
                x = pd.concat([x0[ticker].rename(x_var_name0),
                               x1[ticker].rename(x_var_name1),
                               x2[ticker].rename(x_var_name2),
                               x3[ticker].rename(x_var_name3)], axis=1)
            else:
                x = pd.concat([x1[ticker].rename(x_var_name1),
                               x2[ticker].rename(x_var_name2),
                               x3[ticker].rename(x_var_name3)], axis=1)

            # keep last obs for prediction
            fitted_model = qis.fit_ols(x=x.iloc[:-1, :].to_numpy(), y=y[ticker].iloc[:-1].to_numpy(), order=1, fit_intercept=False)
            actual_change = y[ticker].iloc[-1]
            x_ts = x.iloc[-1, :].to_numpy()
            pred_t = {}
            total_pred = 0.0
            for idx, x_t in enumerate(x_ts):
                pred_x = fitted_model.params[idx] * x_t
                total_pred += pred_x
                pred_t[x_names[idx]] = pred_x
            pred_t['predicted'] = total_pred
            pred_t['actual'] = actual_change
            pred_t['residual'] = actual_change - total_pred
            pred_t['residual %'] = total_pred / actual_change
            pred_t['r2'] = fitted_model.rsquared
            predictions[ticker] = pd.Series(pred_t)
            fitted_models[ticker] = fitted_model

        predictions = pd.DataFrame.from_dict(predictions, orient='index')

        return predictions, fitted_models


@njit
def compute_realized_pnl(prices: np.ndarray,
                         units: np.ndarray
                         ) -> Tuple[np.ndarray, ...]:
    """
    pnl for long only positions, computes avg entry price and pnl by instrument
    """
    avg_costs = np.zeros_like(prices)
    realized_pnl = np.zeros_like(prices)
    mtm_pnl = np.zeros_like(prices)
    trades = np.zeros_like(prices)
    for idx, (price1, unit1) in enumerate(zip(prices, units)):
        if idx == 0:
            avg_costs[idx] = np.where(np.greater(unit1, 1e-16), price1, 0.0)
        else:
            unit0 = units[idx - 1]
            avg_costs0 = avg_costs[idx - 1]
            delta = unit1 - unit0
            is_purchase = np.greater(delta, 1e-16)
            is_sell = np.less(delta, -1e-16)
            realized_pnl[idx] = np.where(is_sell, -delta * (price1 - avg_costs0), 0.0)
            avg_costs[idx] = np.where(is_purchase, np.true_divide(delta * price1 + unit0 * avg_costs0, unit1),
                                      avg_costs0)
            mtm_pnl[idx] = unit0 * (price1 - avg_costs0) - realized_pnl[idx]
            trades[idx] = delta
    return avg_costs, realized_pnl, mtm_pnl, trades


class AllocationType(EnumMap):
    EW = 1
    FIXED_WEIGHTS = 2
    ERC = 3
    ERC_ALT = 4


@dataclass
class PortfolioInput:
    """
    define data inputs for portfolio construction
    """
    name: str
    weights: Union[np.ndarray, pd.DataFrame, Dict[str, float]]
    prices: pd.DataFrame = None  # mandatory but we set none for enumarators
    allocation_type: AllocationType = AllocationType.FIXED_WEIGHTS
    time_period: TimePeriod = None
    rebalancing_freq: str = 'QE'
    freq_regime: str = 'ME'
    returns_freq: str = 'ME'
    ewm_lambda: float = 0.92
    target_vol: float = None

    def update(self, new: Dict[Any, Any]):
        for key, value in new.items():
            if hasattr(self, key):
                setattr(self, key, value)



"""
compute split overnight and intraday returns
"""
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Tuple
import yfinance as yf
import qis


def compute_returns(ticker: str = 'SPY', time_period: qis.TimePeriod = None) -> Tuple[pd.Series, pd.Series, pd.Series]:
    ohlc_data = yf.download(tickers=ticker, start=None, end=None, ignore_tz=True)
    if time_period is not None:
        ohlc_data = time_period.locate(ohlc_data)
    # need to adjust open price for dividends
    adjustment_ratio = ohlc_data['Close'] / ohlc_data['Close']
    adjusted_open = adjustment_ratio.multiply(ohlc_data['Open'])
    adjusted_close = ohlc_data['Close']
    overnight_return = adjusted_open.divide(adjusted_close.shift(1)) - 1.0
    intraday_return = adjusted_close.divide(adjusted_open) - 1.0
    close_to_close_return = adjusted_close.divide(adjusted_close.shift(1)) - 1.0
    return overnight_return.rename('Overnight'), intraday_return.rename('Intraday'), close_to_close_return.rename('Close-to-Close')


def plot_split_returns(ticker: str = 'SPY',
                       time_period: qis.TimePeriod = None,
                       is_check_total: bool = False,
                       ax: plt.Subplot = None
                       ) -> None:
    overnight_return, intraday_return, close_to_close_return = compute_returns(ticker=ticker, time_period=time_period)

    if is_check_total:
        cum_performance = pd.concat([close_to_close_return.rename('Close-to-Close'),
                                     overnight_return.add(intraday_return).rename('Overnight+Intraday')
                                     ], axis=1).cumsum(0)
    else:
        cum_performance = pd.concat([overnight_return, intraday_return], axis=1).cumsum(0)

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(15, 8), tight_layout=True)
    qis.plot_time_series(df=cum_performance,
                         var_format='{:,.0%}',
                         title=f"{ticker} Performance",
                         x_date_freq='YE',
                         date_format='%Y',
                         framealpha=0.9,
                         legend_stats=qis.LegendStats.LAST,
                         ax=ax)


# tickers = ['SPY', 'QQQ', 'GLD', 'TLT', 'HYG', 'MSFT']
tickers = ['AAPL', 'MSFT', 'NVDA', 'AMZN', 'META', 'GOOG']
time_period = qis.TimePeriod('31Dec2004', None)

with sns.axes_style("darkgrid"):
    fig, axs = plt.subplots(2, len(tickers)//2, figsize=(15, 8), tight_layout=True)

for ticker, ax in zip(tickers, qis.to_flat_list(axs)):
    plot_split_returns(ticker=ticker, time_period=time_period, ax=ax)

plt.show()



"""
quantile-quantile plot
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import stats as stats
from statsmodels import api as sm
from typing import List, Union, Tuple, Optional
from enum import Enum

# qis
import qis.perfstats.returns as ret
import qis.plots.utils as put
import qis.perfstats.desc_table as dsc


def plot_qq(df: Union[pd.DataFrame, pd.Series],
            colors: List[str] = None,
            markers: List[str] = None,
            legend_loc: str = 'upper left',
            var_format: str = '{:.2f}',
            is_drop_na: bool = True,
            fontsize: int = 10,
            markersize: int = 2,
            title: str = None,
            xlabel: str = 'Theoretical quantiles',
            ylabel: str = 'Empirical quantiles',
            desc_table_type: dsc.DescTableType = dsc.DescTableType.SHORT,
            legend_stats: put.LegendStats = put.LegendStats.NONE,
            x_limits: Tuple[Optional[float], Optional[float]] = None,
            y_limits: Tuple[Optional[float], Optional[float]] = None,
            ax: plt.Subplot = None,
            **kwargs
            ) -> plt.Figure:

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if isinstance(df, pd.Series):
        df = df.to_frame()
        line = 'q'
    else:
        line= None

    if colors is None:
         colors = put.get_n_colors(n=len(df.columns), **kwargs)

    if markers is None:
        markers = len(df.columns) * ['o']

    for idx, column in enumerate(df.columns):
        data0 = df[column]
        if is_drop_na:
            data0 = data0.dropna()
        sm.qqplot(data0, stats.norm, fit=True, line=line, ax=ax,  fmt=colors[idx],
                  markerfacecolor=colors[idx], markeredgecolor=colors[idx], marker=markers[idx],
                  markersize=markersize)
    if line is None:
        sm.qqline(ax, line='45', fmt='-', color='red')

    if desc_table_type != dsc.DescTableType.NONE:
        stats_table = dsc.compute_desc_table(df=df,
                                         desc_table_type=desc_table_type,
                                         var_format=var_format)
        put.set_legend_with_stats_table(stats_table=stats_table,
                                        ax=ax,
                                        colors=colors,
                                        legend_loc=legend_loc,
                                        fontsize=fontsize,
                                        **kwargs)
    else:
        legend_labels = put.get_legend_lines(data=df,
                                             legend_stats=legend_stats,
                                             var_format=var_format)
        put.set_legend(ax=ax,
                       labels=legend_labels,
                       colors=colors,
                       legend_loc=legend_loc,
                       fontsize=fontsize,
                       **kwargs)

    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)
    put.set_ax_ticks_format(ax=ax, xvar_format=var_format, yvar_format=var_format, fontsize=fontsize, **kwargs)
    put.set_spines(ax=ax, **kwargs)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)
    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize)

    return fig


def plot_xy_qq(x: pd.Series,
               y: pd.Series,
               colors: List[str] = None,
               markers: List[str] = None,
               labels: List[str] = None,
               legend_loc: str = 'upper left',
               is_drop_na: bool = True,
               ax: plt.Subplot = None,
               **kwargs
               ) -> plt.Figure:

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if colors is None:
         colors = put.get_n_colors(n=1)

    if is_drop_na:
        x = x.dropna()
        y = y.dropna()
    x = x.to_numpy()
    y = y.to_numpy()

    qs = np.linspace(0, 1, min(len(x), len(y)))

    x_qs = np.quantile(x, qs)
    y_qs = np.quantile(y, qs)
    ax.scatter(x_qs, y_qs, c=colors[0])

    sm.qqline(ax, line='45', fmt='k--')

    put.set_legend(ax=ax,
                   labels=labels,
                   colors=colors,
                   legend_loc=legend_loc,
                   **kwargs)

    put.set_ax_xy_labels(ax=ax, **kwargs)
    put.set_ax_ticks_format(ax=ax, **kwargs)

    return fig


class UnitTests(Enum):
    RETURNS = 1
    XY_PLOT = 2


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    df = ret.to_returns(prices=prices, drop_first=True)

    if unit_test == UnitTests.RETURNS:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        global_kwargs = dict(fontsize=8, linewidth=0.5, weight='normal', markersize=1)

        plot_qq(df=df,
                desc_table_type=dsc.DescTableType.SKEW_KURTOSIS,
                ax=ax,
                **global_kwargs)

    elif unit_test == UnitTests.XY_PLOT:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        global_kwargs = dict(fontsize=8, linewidth=0.5, weight='normal', markersize=1)
        plot_xy_qq(x=df.iloc[:, 1],
                   y=df.iloc[:, 0],
                   ax=ax,
                   **global_kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.RETURNS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)